from math import ceil, floor, log10
import datetime as dt
import numpy as np
from PyQt5.QtGui import QFontMetrics
import pyqtgraph as pg


class TimeAxisItem(pg.AxisItem):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setPen('white')
        self.StartTime = None
        self.EnableStartTime = False


    def setLogMode(self, *args, **kwargs):
        # no log mode!
        pass


    def setStartTime(self, time):
        """ Set time of first data element.

        Parameters
        ----------
        time: datetime or None
            A datetime object for the data and time of the first data element. 
        """
        self.StartTime = time
        self.enableAutoSIPrefix(self.StartTime is None or
                                not self.EnableStartTime)


    def enableStartTime(self, enable):
        """ Enable addition of start time to tick labels.

        Parameters
        ----------
        enable: bool
            If True enable addition of start time to tick labels.
        """
        self.EnableStartTime = enable
        self.enableAutoSIPrefix(self.StartTime is None or
                                not self.EnableStartTime)


    def tickSpacing(self, minVal, maxVal, size):
        diff = abs(maxVal - minVal)
        if diff == 0:
            return []

        # estimate width of xtick labels:
        xwidth = QFontMetrics(self.font()).averageCharWidth()
        if self.StartTime and self.EnableStartTime:
            nx = 8
        elif maxVal < 1.0:
            nx = 0
        elif maxVal >= 3600:
            nx = 8
        elif maxVal >= 60:
            nx = 5
        else:
            nx = 2
        spacing = diff/5
        if spacing < 0.00001:
            nx += 7
        elif spacing < 0.0001:
            nx += 6
        elif spacing < 0.001:
            nx += 5
        elif spacing < 1.0:
            nx += 4
        nx += 4

        # minimum spacing:
        max_ticks = max(2, int(size / (nx*xwidth)))
        min_spacing = diff / max_ticks
        p10unit = 10 ** floor(log10(min_spacing))

        # major ticks:
        factors = [1.0, 2.0, 5.0, 10.0, 20.0, 50.0, 100.0]
        for fac in factors:
            spacing = fac * p10unit
            if spacing >= min_spacing:
                break

        # minor ticks:
        factors = [100.0, 10.0, 1.0, 0.1]
        for fac in factors:
            minor_spacing = fac * p10unit
            if minor_spacing < spacing:
                break
            
        return [(spacing, 0), (minor_spacing, 0)]

    
    def tickStrings(self, values, scale, spacing):
        if len(values) == 0:
            return []
        
        if scale > 1:
            self.setLabel('Time', units='s')
            return [f'{v*scale:.5g}' for v in values]

        if (self.StartTime and self.EnableStartTime) or np.max(values) > 3600:
            self.setLabel('Time (h:m:s)', units=None)
            fs = '{hours:.0f}:{mins:02.0f}:{secs:02.0f}'
        elif np.max(values) > 60:
            self.setLabel('Time (m:s)', units=None)
            fs = '{mins:.0f}:{secs:02.0f}'
        else:
            self.setLabel('Time', units='s')
            fs = '{secs:.0f}'
        if spacing < 1:
            fs += '.{micros}'
        
        basetime = dt.datetime(1, 1, 1, 0, 0, 0, 0)
        if self.StartTime and self.EnableStartTime:
            basetime = self.StartTime
        vals = []
        for time in values:
            t = basetime + dt.timedelta(seconds=time)
            if spacing < 0.00001:
                micros = f'{1.0*t.microsecond:06.0f}'
            elif spacing < 0.0001:
                micros = f'{0.1*t.microsecond:05.0f}'
            elif spacing < 0.001:
                micros = f'{0.01*t.microsecond:04.0f}'
            else:
                micros = f'{0.001*t.microsecond:03.0f}'
            time = dict(hours=t.hour, mins=t.minute, secs=t.second,
                        micros=micros)
            vals.append(fs.format(**time))
        return vals



import os
from copy import deepcopy
from math import fabs, ceil, floor, log, log10
import datetime as dt
import numpy as np
from scipy.signal import butter, sosfiltfilt
import xml.dom.minidom
try:
    from PyQt5.QtCore import Signal
except ImportError:
    from PyQt5.QtCore import pyqtSignal as Signal
from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QCursor, QKeySequence
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QScrollArea
from PyQt5.QtWidgets import QAction, QMenu, QToolBar, QComboBox
from PyQt5.QtWidgets import QCheckBox, QSpinBox
from PyQt5.QtWidgets import QLabel, QSizePolicy, QTableView
from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QFileDialog
from PyQt5.QtWidgets import QAbstractItemView, QGraphicsRectItem
import pyqtgraph as pg
from audioio import fade
from audioio import get_datetime, update_starttime
from audioio import bext_history_str, add_history
from thunderlab.dataloader import DataLoader
from thunderlab.datawriter import available_formats, write_data
from .version import __version__, __year__
from .fulltraceplot import FullTracePlot, secs_to_str
from .oscillogramplot import OscillogramPlot
from .spectrumplot import SpectrumPlot
from .traceitem import TraceItem
from .specitem import SpecItem
from .markerdata import colors, MarkerLabel, MarkerLabelsModel
from .markerdata import MarkerData, MarkerDataModel


pg.setConfigOption('useNumba', True)


def marker_tip(x, y, data):
    s = ''
    if data:
        s += data + '\n'
    s += 'time=' + secs_to_str(x)
    return s


class DataBrowser(QWidget):

    color_maps = [
                  'CET-R4',   # jet
                  'CET-L8',   # blue-pink-yellow
                  'CET-L16',  # black-blue-green-white
                  'CET-CBL2', # black-blue-yellow-white
                  'CET-L1',   # black-white
                  #pg.colormap.get('CET-L1').reverse(),   # white-black
                  'CET-L3',   # inferno
                  ]
    # see https://colorcet.holoviz.org/
    # and pyqtgraph.colormap module for useful functions.

    zoom_region = 0
    play_region = 1
    save_region = 2
    ask_region = 3
    
    sigTimesChanged = Signal(object, object, object)
    sigAmplitudesChanged = Signal(object, object)
    sigFrequenciesChanged = Signal(object, object)
    sigResolutionChanged = Signal()
    sigColorMapChanged = Signal()
    sigFilterChanged = Signal()
    sigPowerChanged = Signal()
    sigAudioChanged = Signal(object, object, object)

    
    def __init__(self, file_path, channels, audio,
                 acts, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # actions of main window:
        self.acts = acts

        # data:
        self.file_path = file_path
        self.channels = channels
        self.data = None
        self.rate = None
        self.tmax = 0.0
        self.meta_data = {}

        self.show_channels = None
        self.current_channel = 0
        self.selected_channels = []
        
        self.trace_fracs = {0: 1, 1: 1, 2: 0.5, 3: 0.25, 4: 0.15}

        self.region_mode = DataBrowser.ask_region
        
        # view:
        self.toffset = 0.0
        self.twindow = 10.0

        self.setting = False
        
        self.grids = 0
        self.show_traces = True
        self.show_specs = 4
        self.show_cbars = False
        self.show_fulldata = True
        
        # auto scroll:
        self.scroll_step = 0.0
        self.scroll_timer = QTimer(self)
        self.scroll_timer.timeout.connect(self.scroll_further)

        # audio:
        self.audio = audio
        self.audio_timer = QTimer(self)
        self.audio_timer.timeout.connect(self.mark_audio)
        self.audio_time = 0.0
        self.audio_use_heterodyne = False
        self.audio_heterodyne_freq = 40000.0
        self.audio_rate_fac = 1.0
        self.audio_tmax = 0.0
        self.audio_markers = [] # vertical lines showing position while playing

        # window:
        self.vbox = QVBoxLayout(self)
        self.vbox.setContentsMargins(0, 0, 0, 0)
        self.vbox.setSpacing(0)
        self.setEnabled(False)
        self.toolbar = None
        self.audiofacw = None
        self.nfftw = None

        # cross hair:
        self.xpos_action = None
        self.ypos_action = None
        self.zpos_action = None
        self.cross_hair = False
        self.marker_ax = None
        self.marker_time = 0
        self.marker_ampl = 0
        self.marker_freq = 0
        self.marker_power = 0
        self.marker_channel = None
        self.prev_time = 0
        self.prev_ampl = 0
        self.prev_freq = 0
        self.prev_power = 0
        self.prev_channel = None
        self.delta_time = None
        self.delta_ampl = None
        self.delta_freq = None
        self.delta_power = None
        self.marker_data = MarkerData()
        self.marker_model = MarkerDataModel(self.marker_data)
        self.marker_labels = []
        self.marker_labels.append(MarkerLabel('start', 's', 'yellow'))
        self.marker_labels.append(MarkerLabel('end', 'e', 'blue'))
        self.marker_labels_model = MarkerLabelsModel(self.marker_labels,
                                                     self.acts)
        self.marker_orig_acts = []
        
        # plots:
        self.color_map = 0  # index into color_maps
        self.figs = []      # all GraphicsLayoutWidgets - one for each channel
        self.borders = []
        self.sig_proxies = []
        # nested lists (channel, panel):
        self.axs  = []      # all plots
        self.axts = []      # plots with time axis
        self.axys = []      # plots with amplitude axis
        self.axfxs = []     # plots with x-frequency axis
        self.axfys = []     # plots with y-frequency axis
        self.axgs = []      # plots with grids
        # lists with one plot per channel:
        self.axtraces = []  # trace plots
        self.axspacers = [] # spacer between trace and spectrogram
        self.axspecs = []   # spectrogram plots
        self.traces = []    # traces
        self.specs = []     # spectrograms
        self.cbars = []     # color bars
        self.trace_labels = [] # labels on traces
        self.trace_region_labels = [] # regions with labels on traces
        self.spec_labels = []  # labels on spectrograms
        self.spec_region_labels = [] # regions with labels on spectrograms
        self.datafig = None   # full traces

        
    def __del__(self):
        if not self.data is None:
            self.data.close()

        
    def open(self, gui, unwrap, unwrap_clip):
        if not self.data is None:
            self.data.close()
        try:
            self.data = DataLoader(self.file_path, 60.0, 10.0)
        except IOError:
            self.data = None
            return
        self.data.set_unwrap(unwrap, unwrap_clip, False, self.data.unit)
        self.file_path = self.data.filepath
        self.rate = self.data.samplerate
        self.marker_data.file_path = self.file_path

        self.toffset = 0.0
        self.twindow = 10.0
        self.tmax = len(self.data)/self.rate
        if self.twindow > self.tmax:
            self.twindow = self.tmax

        if self.show_channels is None:
            if len(self.channels) == 0:
                self.show_channels = list(range(self.data.channels))
            else:
                self.show_channels = [c for c in self.channels if c < self.data.channels]
        else:
            self.show_channels = [c for c in self.show_channels if c < self.data.channels]
        if len(self.show_channels) == 0:
            self.show_channels = [0]
        
        self.current_channel = self.show_channels[0]
        self.selected_channels = list(range(self.data.channels))

        # load data:
        self.meta_data = dict(Format=self.data.format_dict())
        self.meta_data.update(self.data.metadata())
        starttime = get_datetime(self.meta_data)
        locs, labels = self.data.markers()
        self.marker_data.set_markers(locs, labels, self.rate)
        if len(labels) > 0:
            lbls = np.unique(labels[:,0])
            for i, l in enumerate(lbls):
                self.marker_labels.append(MarkerLabel(l, l[0].lower(),
                                list(colors.keys())[i % len(colors.keys())]))
        self.data[0,:]     # load first frame

        self.figs = []     # all GraphicsLayoutWidgets - one for each channel
        self.borders = []
        self.sig_proxies = []
        # nested lists (channel, panel):
        self.axs  = []      # all plots
        self.axts = []      # plots with time axis
        self.axys = []      # plots with amplitude axis
        self.axfxs = []     # plots with x-frequency axis
        self.axfys = []     # plots with y-frequency axis
        self.axgs = []      # plots with grids
        # lists with one plot per channel:
        self.axtraces = []  # trace plots
        self.axspacers = [] # spacer between trace and spectrogram
        self.axspecs = []   # spectrogram plots
        self.traces = []    # traces
        self.trace_labels = [] # labels on traces
        self.trace_region_labels = [] # regions with labels on traces
        self.specs = []     # spectrograms
        self.spec_labels = []  # labels on spectrograms
        self.spec_region_labels = [] # regions with labels on spectrograms
        self.cbars = []     # color bars
        self.audio_markers = [] # vertical line showing position while playing
        # font size:
        xwidth = self.fontMetrics().averageCharWidth()
        xwidth2 = xwidth/2
        for c in range(self.data.channels):
            self.axs.append([])
            self.axts.append([])
            self.axys.append([])
            self.axfxs.append([])
            self.axfys.append([])
            self.axgs.append([])
            self.audio_markers.append([])
            
            # one figure per channel:
            fig = pg.GraphicsLayoutWidget()
            fig.setBackground(None)
            fig.ci.layout.setContentsMargins(xwidth2, xwidth2, xwidth2, xwidth2)
            fig.ci.layout.setVerticalSpacing(0)
            fig.ci.layout.setHorizontalSpacing(xwidth)
            fig.setVisible(c in self.show_channels)

            self.vbox.addWidget(fig)
            self.figs.append(fig)
            
            # border:
            border = QGraphicsRectItem()
            border.setZValue(-1000)
            border.setPen(pg.mkPen('#aaaaaa', width=xwidth+1))
            fig.scene().addItem(border)
            fig.sigDeviceRangeChanged.connect(self.update_borders)
            self.borders.append(border)
            
            # spectrogram:
            # takes a long time:
            spec = SpecItem(self.data, self.rate, c, 256, 0.5)
            self.specs.append(spec)
            axs = SpectrumPlot(c, xwidth, starttime, spec.fmax)
            axs.addItem(spec)
            labels = []
            for l in self.marker_labels:
                label = pg.ScatterPlotItem(size=10, pen=pg.mkPen(None),
                                           brush=pg.mkBrush(l.color))
                axs.addItem(label)
                labels.append(label)
            self.spec_labels.append(labels)
            self.spec_region_labels.append([])
            axs.setLimits(xMin=0, xMax=self.tmax,
                         minXRange=10/self.rate, maxXRange=self.tmax)
            axs.setXRange(self.toffset, self.toffset + self.twindow)
            axs.sigXRangeChanged.connect(self.update_times)
            axs.setYRange(self.specs[c].f0, self.specs[c].f1)
            axs.sigYRangeChanged.connect(self.update_frequencies)
            axs.sigSelectedRegion.connect(self.region_menu)
            axs.sigUpdateFilter.connect(self.update_filter)
            axs.getViewBox().init_zoom_history()
            self.audio_markers[-1].append(axs.vmarker)
            fig.addItem(axs, row=0, col=0)
            
            # color bar:
            cbar = pg.ColorBarItem(colorMap=self.color_maps[self.color_map],
                                   interactive=True,
                                   rounding=1, limits=(-200, 20))
            cbar.setLabel('right', 'Power (dB)')
            cbar.getAxis('right').setTextPen('black')
            cbar.getAxis('right').setWidth(6*xwidth)
            cbar.setLevels([spec.zmin, spec.zmax])
            cbar.setImageItem(spec)
            cbar.sigLevelsChanged.connect(self.update_power)
            cbar.setVisible(self.show_cbars)
            self.cbars.append(cbar)
            fig.addItem(cbar, row=0, col=1)
            spec.setCBar(cbar)
            self.axts[-1].append(axs)
            self.axfys[-1].append(axs)
            self.axgs[-1].append(axs)
            self.axs[-1].append(axs)
            self.axspecs.append(axs)
            
            # spacer:
            axsp = fig.addLayout(row=1, col=0)
            axsp.setContentsMargins(0, 0, 0, 0)
            self.axspacers.append(axsp)
            
            # trace plot:
            trace = TraceItem(self.data, self.rate, c)
            self.traces.append(trace)
            # takes some time:
            axt = OscillogramPlot(c, xwidth, starttime, self.data.channels > 4)
            axt.addItem(trace)
            labels = []
            for l in self.marker_labels:
                label = pg.ScatterPlotItem(size=10, hoverSize=20,
                                           hoverable=True,
                                           pen=pg.mkPen(None),
                                           brush=pg.mkBrush(l.color))
                axt.addItem(label)
                labels.append(label)
            self.trace_labels.append(labels)
            self.trace_region_labels.append([])
            axt.getAxis('bottom').showLabel(c == self.show_channels[-1])
            axt.getAxis('bottom').setStyle(showValues=(c == self.show_channels[-1]))
            axt.setLimits(xMin=0, xMax=self.tmax,
                          minXRange=10/self.rate, maxXRange=self.tmax)
            if np.isfinite(self.data.ampl_min) and np.isfinite(self.data.ampl_max):
                axt.setLimits(yMin=self.data.ampl_min, yMax=self.data.ampl_max,
                              minYRange=1/2**16,
                              maxYRange=self.data.ampl_max - self.data.ampl_min)

            axt.setXRange(self.toffset, self.toffset + self.twindow)
            axt.sigXRangeChanged.connect(self.update_times)
            axt.setYRange(self.traces[c].ymin, self.traces[c].ymax)
            axt.sigYRangeChanged.connect(self.update_amplitudes)
            axt.sigSelectedRegion.connect(self.region_menu)
            axt.getViewBox().init_zoom_history()
            self.audio_markers[-1].append(axt.vmarker)
            fig.addItem(axt, row=2, col=0)
            self.axts[-1].append(axt)
            self.axys[-1].append(axt)
            self.axgs[-1].append(axt)
            self.axs[-1].append(axt)
            self.axtraces.append(axt)

            proxy = pg.SignalProxy(fig.scene().sigMouseMoved, rateLimit=60,
                                   slot=lambda x, c=c: self.mouse_moved(x, c))
            self.sig_proxies.append(proxy)
            proxy = pg.SignalProxy(fig.scene().sigMouseClicked, rateLimit=60,
                                   slot=lambda x, c=c: self.mouse_clicked(x, c))
            self.sig_proxies.append(proxy)
            
        self.set_times()
        
        # tool bar:
        self.toolbar = QToolBar()
        self.toolbar.addAction(self.acts.skip_backward)
        self.toolbar.addAction(self.acts.seek_backward)
        self.toolbar.addAction(self.acts.seek_forward)
        self.toolbar.addAction(self.acts.skip_forward)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.acts.play_window)
        self.audiofacw = QComboBox(self)
        self.audiofacw.setToolTip('Audio time expansion factor')
        self.audiofacw.addItems(['0.1', '0.2', '0.5', '1', '2', '5', '10', '20', '50', '100'])
        self.audiofacw.setEditable(False)
        self.audiofacw.setCurrentText(f'{self.audio_rate_fac:g}')
        self.audiofacw.currentTextChanged.connect(lambda s: self.set_audio(rate_fac=float(s)))
        self.toolbar.addWidget(self.audiofacw)
        self.audiohetfw = QSpinBox(self)
        self.audiohetfw.setToolTip('Audio heterodyne frequency')
        self.audiohetfw.setRange(10, 100)
        self.audiohetfw.setSingleStep(5)
        self.audiohetfw.setSuffix('kHz')
        self.audiohetfw.setValue(int(self.audio_heterodyne_freq/1000))
        self.audiohetfw.valueChanged.connect(lambda v: self.set_audio(heterodyne_freq=1000*v))
        if self.data.samplerate > 50000:
            self.toolbar.addWidget(self.audiohetfw)
        else:
            self.audiohetfw.setVisible(False)
        if self.data.samplerate > 50000:
            self.toolbar.addAction(self.acts.use_heterodyne)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.acts.zoom_home)
        self.toolbar.addAction(self.acts.zoom_back)
        self.toolbar.addAction(self.acts.zoom_forward)
        self.toolbar.addSeparator()
        self.nfftw = QComboBox(self)
        self.nfftw.setToolTip('NFFT (R, Shift+R)')
        self.nfftw.addItems([f'{2**i}' for i in range(4, 16)])
        self.nfftw.setEditable(False)
        self.nfftw.setCurrentText(f'{self.specs[self.current_channel].nfft}')
        self.nfftw.currentTextChanged.connect(lambda s: self.set_resolution(nfft=int(s)))
        self.toolbar.addWidget(self.nfftw)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(QLabel('Channel:'))
        for c in range(max(self.data.channels, len(self.acts.channels))):
            gui.set_channel_action(c, self.data.channels,
                                   c in self.show_channels,
                                   gui.browser() is self)
            if c < self.data.channels:
                self.toolbar.addAction(self.acts.channels[c])
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.toolbar.addWidget(spacer)
        self.xpos_action = self.toolbar.addAction('xpos')
        self.xpos_action.setVisible(False)
        self.toolbar.widgetForAction(self.xpos_action).setFixedWidth(20*xwidth)
        self.ypos_action = self.toolbar.addAction('ypos')
        self.ypos_action.setVisible(False)
        self.toolbar.widgetForAction(self.ypos_action).setFixedWidth(10*xwidth)
        self.zpos_action = self.toolbar.addAction('zpos')
        self.zpos_action.setVisible(False)
        self.toolbar.widgetForAction(self.zpos_action).setFixedWidth(10*xwidth)
        self.vbox.addWidget(self.toolbar)
        
        # full data:
        self.datafig = FullTracePlot(self.data, self.rate, self.axtraces)
        self.vbox.addWidget(self.datafig)

        self.setEnabled(True)
        self.adjust_layout(self.width(), self.height())

        # add marker data to plot:
        labels = [l.label for l in self.marker_labels]
        for t1, ddt, ls, ts in zip(self.marker_data.times,
                                  self.marker_data.delta_times,
                                  self.marker_data.labels,
                                  self.marker_data.texts):
            lidx = labels.index(ls)
            for c, tl in enumerate(self.trace_labels):
                ds = ts if ts else ls
                t0 = t1 - ddt
                idx0 = int(t0*self.rate)
                idx1 = int(t1*self.rate)
                if ddt > 0:
                    region = pg.LinearRegionItem((t0, t1),
                                                 orientation='vertical',
                                                 pen=pg.mkPen(self.marker_labels[lidx].color),
                                                 brush=pg.mkBrush(self.marker_labels[lidx].color),
                                                 movable=False,
                                                 span=(0.02, 0.05))
                    region.setZValue(-10)
                    self.axtraces[c].addItem(region)
                    #text = pg.TextItem(ds, color='green', anchor=(0, 0))
                    #text.setPos(t0, 0)
                    #self.axtraces[c].addItem(text)
                    self.trace_region_labels[c].append(region)
                else:
                    tl[lidx].addPoints((t1,), (self.data[idx1, c],), data=(ds,), tip=marker_tip)
            for c, sl in enumerate(self.spec_labels):
                if ddt > 0:
                    # TODO: self.spec_region_labels
                    sl[lidx].addPoints((t0, t1), (0.0, 0.0),
                                       data=(f'start: {ds}', f'end: {ds}'))
                else:
                    sl[lidx].addPoints((t1,), (0.0,), data=(ds,), tip=marker_tip)


    def show_metadata(self):
        
        def format_dict(md, level):
            mdtable = ''
            for k in md:
                pads = ''
                if level > 0:
                    pads = f' style="padding-left: {level*30:d}px;"'
                if isinstance(md[k], dict):
                    # new section:
                    if level == 0:
                        mdtable += f'<tr><td colspan=2><font size="+1"><b>{k}:</b></font></td></tr>'
                    else:
                        mdtable += f'<tr><td colspan=2{pads}><b>{k}:</b></td></tr>'
                    mdtable += format_dict(md[k], level+1)
                    if level == 0:
                        mdtable += '<tr><td colspan=2></td></tr>'
                else:
                    # key-value pair:
                    value = md[k]
                    if isinstance(value, (list, tuple)):
                        value = ', '.join([f'{v}' for v in value])
                    else:
                        value = f'{value}'
                    value = value.replace('\r\n', '\n')
                    value = value.replace('\r', '\n')
                    value = value.replace('\n', '<br>')
                    mdtable += f'<tr><td{pads}><b>{k}</b></td><td>{value}</td></tr>'
            return mdtable

        w = xwidth = self.fontMetrics().averageCharWidth()
        mdtable = f'<style>td {{padding: 0 {w}px 0 0; }}</style><table>'
        mdtable += format_dict(self.meta_data, 0)
        mdtable += '</table>'
        dialog = QDialog(self)
        dialog.setWindowTitle('Meta data')
        vbox = QVBoxLayout()
        dialog.setLayout(vbox)
        label = QLabel(mdtable)
        label.setTextInteractionFlags(Qt.TextSelectableByMouse);
        scrollarea = QScrollArea()
        scrollarea.setWidget(label)
        vbox.addWidget(scrollarea)
        buttons = QDialogButtonBox(QDialogButtonBox.Close)
        buttons.rejected.connect(dialog.reject)
        vbox.addWidget(buttons)
        dialog.show()


    def set_cross_hair(self, checked):
        self.cross_hair = checked
        if self.cross_hair:
            # disable existing key shortcuts:
            self.marker_orig_acts = []
            for l in self.marker_labels:
                ks = QKeySequence(l.key_shortcut)
                for a in dir(self.acts):
                    act = getattr(self.acts, a)
                    if isinstance(act, QAction) and act.shortcut() == ks:
                        self.marker_orig_acts.append((act.shortcut(), act))
                        act.setShortcut(QKeySequence())
                        break
            # setup marker actions:            
            for l in self.marker_labels:
                if l.action is None:
                    l.action = QAction(l.label, self)
                    l.action.triggered.connect(lambda x, label=l.label: self.store_marker(label))
                    self.addAction(l.action)
                l.action.setShortcut(l.key_shortcut)
                l.action.setEnabled(True)
            for axts in self.axts:
                for ax in axts:
                    ax.xline.setVisible(True)
            for axys in self.axys:
                for ax in axys:
                    ax.yline.setVisible(True)
            for axfys in self.axfys:
                for ax in axfys:
                    ax.yline.setVisible(True)
        else:
            self.xpos_action.setVisible(False)
            self.ypos_action.setVisible(False)
            self.zpos_action.setVisible(False)
            for axts in self.axts:
                for ax in axts:
                    ax.xline.setVisible(False)
            for axys in self.axys:
                for ax in axys:
                    ax.yline.setVisible(False)
            for axfys in self.axfys:
                for ax in axfys:
                    ax.yline.setVisible(False)
            self.clear_marker()
            # disable marker actions:
            for l in self.marker_labels:
                l.action.setEnabled(False)
            # restore key shortcuts:
            for key, act in self.marker_orig_acts:
                act.setShortcuts(key)
            self.marker_orig_acts = []


    def clear_marker(self):
        for axs in self.axs:
            for axp in axs:
                if hasattr(axp, 'prev_marker'):
                    axp.prev_marker.clear()
        self.prev_channel = None


    def set_marker(self):
        self.clear_marker()
        if not self.marker_ax is None and not self.marker_time is None:
            if not self.marker_ampl is None:
                self.marker_ax.prev_marker.setData((self.marker_time,),
                                                   (self.marker_ampl,))
            if not self.marker_freq is None:
                self.marker_ax.prev_marker.setData((self.marker_time,),
                                                   (self.marker_freq,))
            # remember:
            self.prev_time = self.marker_time
            self.prev_ampl = self.marker_ampl
            self.prev_freq = self.marker_freq
            self.prev_power = self.marker_power
            self.prev_channel = self.marker_channel

            
    def store_marker(self, label=''):
        self.marker_model.add_data(self.marker_channel,
                                   self.marker_time, self.marker_ampl,
                                   self.marker_freq,
                                   self.marker_power,self.delta_time,
                                   self.delta_ampl, self.delta_freq,
                                   self.delta_power, label)
        # add new label point to scatter plots:
        labels = [l.label for l in self.marker_labels]
        if len(label) > 0 and label in labels and \
           not self.marker_time is None:
            lidx = labels.index(label)
            for c, tl in enumerate(self.trace_labels):
                if c == self.marker_channel and not self.marker_ampl is None:
                    tl[lidx].addPoints((self.marker_time,),
                                      (self.marker_ampl,),
                                       tip=marker_tip)
                else:
                    tidx = int(self.marker_time*self.rate)
                    tl[lidx].addPoints((self.marker_time,),
                                       (self.data[tidx, c],),
                                       tip=marker_tip)
            for c, sl in enumerate(self.spec_labels):
                y = 0.0 if self.marker_freq is None else self.marker_freq
                sl[lidx].addPoints((self.marker_time,), (y,))
                
        
    def mouse_moved(self, evt, channel):
        if not self.cross_hair:
            return
            
        # find axes and position:
        pixel_pos = evt[0]
        self.marker_ax = None
        self.marker_time = None
        self.marker_ampl = None
        self.marker_freq = None
        self.marker_power = None
        self.marker_channel = channel
        for ax in self.axs[channel]:
            if ax.sceneBoundingRect().contains(pixel_pos):
                pos = ax.getViewBox().mapSceneToView(pixel_pos)
                pixel_pos.setX(pixel_pos.x()+1)
                npos = ax.getViewBox().mapSceneToView(pixel_pos)
                if hasattr(ax, 'xline'):
                    ax.xline.setPos(pos.x())
                    # is it time?
                    for axts in self.axts:
                        if ax in axts:
                            self.marker_ax = ax
                            self.marker_time = pos.x()
                            break
                if hasattr(ax, 'yline'):
                    ax.yline.setPos(pos.y())
                    # is it amplitude?
                    for axys in self.axys:
                        if ax in axys:
                            self.marker_ampl = pos.y()
                            break
                    # is it trace amplitude?
                    if ax in self.axtraces:
                        if not self.marker_time is None:
                            trace = self.traces[self.axtraces.index(ax)]
                            self.marker_time, self.marker_ampl = \
                                trace.get_amplitude(self.marker_time,
                                                    pos.y(), npos.x())
                    # is it frequency?
                    for axfys in self.axfys:
                        if ax in axfys:
                            self.marker_freq = pos.y()
                            break
                    # is it spectrogram?
                    if self.marker_time is not None and \
                       self.marker_freq is not None and ax in self.axspecs:
                        spec = self.specs[self.axspecs.index(ax)]
                        fi = int(floor(self.marker_freq/spec.fresolution))
                        ti = int(floor((self.marker_time - spec.offset/spec.rate) / spec.tresolution))
                        if fi < spec.spectrum.shape[0] and \
                           ti < spec.spectrum.shape[1]:
                            self.marker_power = spec.spectrum[fi, ti]
                break
            
        # set cross-hair positions:
        if self.marker_time:
            for axts in self.axts:
                for axt in axts:
                    axt.xline.setPos(self.marker_time)
        if self.marker_ampl:
            for axys in self.axys:
                for axy in axys:
                    axy.yline.setPos(self.marker_ampl)
        if self.marker_freq:
            for axfys in self.axfys:
                for axf in axfys:
                    axf.yline.setPos(self.marker_freq)
                
        # compute deltas:
        self.delta_time = None
        self.delta_ampl = None
        self.delta_freq = None
        self.delta_power = None
        if self.marker_time is not None and \
           self.prev_channel is not None and self.prev_time is not None:
                self.delta_time = self.marker_time - self.prev_time
        if self.marker_ampl is not None and \
           self.prev_channel is not None and self.prev_ampl is not None:
                self.delta_ampl = self.marker_ampl - self.prev_ampl
        if self.marker_freq is not None and \
           self.prev_channel is not None and self.prev_freq is not None:
                self.delta_freq = self.marker_freq - self.prev_freq
        if self.marker_power is not None and \
           self.prev_channel is not None and self.prev_power is not None:
                self.delta_power = self.marker_power - self.prev_power
        
        # report time on toolbar:
        if self.delta_time is not None:
            sign = '-' if self.delta_time < 0 else ''
            s = f'\u0394t={sign}{secs_to_str(fabs(self.delta_time))}'
            if fabs(self.delta_time) > 1e-6:
                s += f': f={1/fabs(self.delta_time):.5g}Hz'
            self.xpos_action.setText(s)
        elif self.marker_time is not None:
            sign = '-' if self.marker_time < 0 else ''
            s = f't={sign}{secs_to_str(fabs(self.marker_time))}'
            self.xpos_action.setText(s)
        else:
            self.xpos_action.setText('')
        # report amplitude or frequency on toolbar:
        if self.delta_ampl is not None:
            s = f'\u0394a={self.delta_ampl:6.3f}'
            self.ypos_action.setText(s)
        elif self.marker_ampl is not None:
            s = f'a={self.marker_ampl:6.3f}'
            self.ypos_action.setText(s)
        elif self.delta_freq is not None:
            s = f'\u0394f={self.delta_freq:4.0f}Hz'
            self.ypos_action.setText(s)
        elif self.marker_freq is not None:
            s = f'f={self.marker_freq:4.0f}Hz'
            self.ypos_action.setText(s)
        else:
            self.ypos_action.setText('')
        # report power on toolbar:
        if self.delta_power is not None:
            s = f'\u0394p={self.delta_power:6.1f}dB'
            self.zpos_action.setText(s)
        elif self.marker_power is not None:
            s = f'p={self.marker_power:6.1f}dB'
            self.zpos_action.setText(s)
        else:
            self.zpos_action.setText('')
        self.xpos_action.setVisible(self.marker_time is not None)
        self.ypos_action.setVisible(self.marker_ampl is not None or
                                    self.marker_freq is not None)
        self.zpos_action.setVisible(self.marker_power is not None)


    def mouse_clicked(self, evt, channel):
        if not self.cross_hair:
            return
        
        # update position:
        self.mouse_moved((evt[0].scenePos(),), channel)

        # store marker positions:
        if (evt[0].button() & Qt.LeftButton) > 0 and \
           (evt[0].modifiers() == Qt.NoModifier or \
            (evt[0].modifiers() & Qt.ShiftModifier) == Qt.ShiftModifier):
            menu = QMenu(self)
            acts = [menu.addAction(self.marker_labels_model.icons[l.color], l.label) for l in self.marker_labels]
            act = menu.exec(QCursor.pos())
            if act in acts:
                idx = acts.index(act)
                self.store_marker(self.marker_labels[idx].label)

        # clear marker:
        if (evt[0].button() & Qt.RightButton) > 0:
            self.clear_marker()
            
        # set marker and remember position:
        if (evt[0].button() & Qt.LeftButton) > 0 and \
           (evt[0].modifiers() & Qt.ControlModifier) == Qt.ControlModifier:
            self.set_marker()

            
    def label_editor(self):
        self.marker_labels_model.set(self.marker_labels)
        self.marker_labels_model.edit(self)
        
        
    def marker_table(self):
        dialog = QDialog(self)
        dialog.setWindowTitle('Audian marker table')
        vbox = QVBoxLayout()
        dialog.setLayout(vbox)
        view = QTableView()
        view.setModel(self.marker_model)
        view.resizeColumnsToContents()
        width = view.verticalHeader().width() + 24
        for c in range(self.marker_model.columnCount()):
            width += view.columnWidth(c)
        dialog.setMaximumWidth(width)
        dialog.resize(width, 2*width//3)
        view.setSelectionMode(QAbstractItemView.ContiguousSelection)
        vbox.addWidget(view)
        buttons = QDialogButtonBox(QDialogButtonBox.Close |
                                   QDialogButtonBox.Save |
                                   QDialogButtonBox.Reset)
        buttons.rejected.connect(dialog.reject)
        buttons.button(QDialogButtonBox.Reset).clicked.connect(self.marker_model.clear)
        buttons.button(QDialogButtonBox.Save).clicked.connect(lambda x: self.marker_model.save(self))
        vbox.addWidget(buttons)
        dialog.show()
            

    def update_borders(self, rect=None):
        for c in range(len(self.figs)):
            self.borders[c].setRect(0, 0, self.figs[c].size().width(),
                                    self.figs[c].size().height())
            self.borders[c].setVisible(c in self.selected_channels)


    def showEvent(self, event):
        if self.data is None:
            return
        self.setting = True
        for c in range(self.data.channels):
            # update time ranges:
            for ax in self.axts[c]:
                ax.setXRange(self.toffset, self.toffset + self.twindow)
            # update amplitude ranges:
            for ax in self.axys[c]:
                ax.setYRange(self.traces[c].ymin, self.traces[c].ymax)
            # update frequency ranges:
            for ax in self.axfys[c]:
                ax.setYRange(self.specs[c].f0, self.specs[c].f1)
            for ax in self.axfxs[c]:
                ax.setXRange(self.specs[c].f0, self.specs[c].f1)
            # update spectrograms:
            self.specs[c].update_spectrum()
        self.setting = False

                
    def resizeEvent(self, event):
        if self.show_channels is None or len(self.show_channels) == 0:
            return
        self.adjust_layout(event.size().width(), event.size().height())
        

    def adjust_layout(self, width, height):
        if self.show_channels is None:
            return
        xwidth = self.fontMetrics().averageCharWidth()
        xheight = self.fontMetrics().ascent()
        # subtract full data plot:
        data_height = 5*xheight//2 if len(self.show_channels) <= 1 else 3*xheight//2
        if not self.show_fulldata:
            data_height = 0
        height -= len(self.show_channels)*data_height
        # subtract toolbar:
        height -= 2*xheight
        bottom_channel = self.show_channels[-1]
        trace_frac = self.trace_fracs[self.show_specs]
        #axis_height = None
        #if self.axtraces[bottom_channel].isVisible():
        #    axis_height = self.axtraces[bottom_channel].getAxis('bottom').height()
        #elif self.axspecs[bottom_channel].isVisible():
        #    axis_height = self.axspecs[bottom_channel].getAxis('bottom').height()
        axis_height = 3.2*xheight
        ntraces = []
        nspecs = []
        nspacer = 0
        for c in self.show_channels:
            if c >= len(self.axspecs) or c >= len(self.axtraces):
                break
            nspecs.append(int(self.axspecs[c].isVisible()))
            ntraces.append(int(self.axtraces[c].isVisible()))
            if self.axspecs[c].isVisible() and self.axtraces[c].isVisible():
                nspacer += 1
        spec_height = (height - len(self.show_channels)*xwidth - nspacer*xwidth - axis_height)/(np.sum(nspecs) + trace_frac*np.sum(ntraces))
        for c, ns, nt in zip(self.show_channels, nspecs, ntraces):
            add_height = axis_height if c == bottom_channel else 0
            self.vbox.setStretch(c, int(ns*spec_height + (nt+ns)*xwidth +
                                        nt*trace_frac*spec_height + add_height))
            t_height = max(0, int(nt*(trace_frac*spec_height + add_height)))
            self.figs[c].ci.layout.setRowFixedHeight(2, t_height)
            self.figs[c].ci.layout.setRowFixedHeight(1, (nt+ns-1)*xwidth)
            s_height = max(0, int(ns*spec_height + (1-nt)*add_height))
            self.figs[c].ci.layout.setRowFixedHeight(0, s_height)
        # fix full data plot:
        if self.datafig is not None:
            self.datafig.update_layout(self.show_channels, data_height)
            self.datafig.setVisible(self.show_fulldata)
        # update:
        for c in self.show_channels:
            self.figs[c].update()
        
            
    def show_xticks(self, channel, show_ticks):
        if self.axtraces[channel].isVisible():
            self.axtraces[channel].getAxis('bottom').showLabel(show_ticks)
            self.axtraces[channel].getAxis('bottom').setStyle(showValues=show_ticks)
            self.axspecs[channel].getAxis('bottom').showLabel(False)
            self.axspecs[channel].getAxis('bottom').setStyle(showValues=False)
        elif self.axspecs[channel].isVisible():
            self.axspecs[channel].getAxis('bottom').showLabel(show_ticks)
            self.axspecs[channel].getAxis('bottom').setStyle(showValues=show_ticks)


    def set_times(self, toffset=None, twindow=None, enable_starttime=None,
                  dispatch=True):
        self.setting = True
        if not toffset is None:
            self.toffset = toffset
        if not twindow is None:
            self.twindow = twindow
        for axs in self.axts:
            for ax in axs:
                if enable_starttime is not None:
                    ax.enableStartTime(enable_starttime)
                if self.isVisible():
                    ax.setXRange(self.toffset, self.toffset + self.twindow)
        self.setting = False
        if dispatch:
            self.sigTimesChanged.emit(self.toffset, self.twindow,
                                      enable_starttime)

            
    def update_times(self, viewbox, trange):
        if self.setting:
            return
        self.toffset = trange[0]
        self.twindow = trange[1] - trange[0]
        self.set_times()
        
        
    def zoom_time_in(self):
        if self.twindow * self.rate >= 20:
            self.twindow *= 0.5
            self.set_times()
        
        
    def zoom_time_out(self):
        if self.toffset + self.twindow < self.tmax:
            self.twindow *= 2.0
            self.set_times()

                
    def time_seek_forward(self):
        if self.toffset + self.twindow < self.tmax:
            self.toffset += 0.5*self.twindow
            self.set_times()

            
    def time_seek_backward(self):
        if self.toffset > 0:
            self.toffset -= 0.5*self.twindow
            if self.toffset < 0.0:
                self.toffset = 0.0
            self.set_times()

                
    def time_forward(self):
        if self.toffset + self.twindow < self.tmax:
            self.toffset += 0.05*self.twindow
            self.set_times()

                
    def time_backward(self):
        axt = None
        for axs in self.axts:
            if len(axs) > 0:
                axt = axs[0]
        if axt is None:
            return
        rect = axt.getViewBox().viewRect()
        toffs = self.toffset
        if toffs > rect.left():
            toffs = rect.left()
        if toffs > 0.0:
            self.toffset = toffs - 0.05*self.twindow
            if self.toffset < 0.0:
                self.toffset = 0.0
            self.set_times()

                
    def time_home(self):
        if self.toffset > 0.0:
            self.toffset = 0.0
            self.set_times()

                
    def time_end(self):
        n2 = np.floor(self.tmax / (0.5*self.twindow))
        toffs = max(0, n2-1)  * 0.5*self.twindow
        if self.toffset < toffs:
            self.toffset = toffs
            self.set_times()

                
    def snap_time(self):
        twindow = 10.0 * 2**np.round(log(self.twindow/10.0)/log(2.0))
        toffset = np.round(self.toffset / (0.5*twindow)) * (0.5*twindow)
        if twindow != self.twindow or toffset != self.toffset:
            self.set_times(toffset, twindow)


    def set_amplitudes(self, ymin=None, ymax=None):
        self.setting = True
        for c in self.selected_channels:
            if not ymin is None:
                self.traces[c].ymin = ymin
            if not ymax is None:
                self.traces[c].ymax = ymax
            if self.isVisible():
                for ax in self.axys[c]:
                    ax.setYRange(self.traces[c].ymin, self.traces[c].ymax)
        self.setting = False

            
    def update_amplitudes(self, viewbox, arange):
        if self.setting:
            return
        self.set_amplitudes(arange[0], arange[1])
        self.sigAmplitudesChanged.emit(arange[0], arange[1])
        

    def zoom_ampl_in(self):
        for c in self.selected_channels:
            self.traces[c].zoom_ampl_in()
        self.set_amplitudes()

        
    def zoom_ampl_out(self):
        for c in self.selected_channels:
            self.traces[c].zoom_ampl_out()
        self.set_amplitudes()
        
        
    def auto_ampl(self):
        for c in self.selected_channels:
            self.traces[c].auto_ampl(self.toffset, self.twindow)
        self.set_amplitudes()

        
    def reset_ampl(self):
        for c in self.selected_channels:
            self.traces[c].reset_ampl()
        self.set_amplitudes()


    def center_ampl(self):
        for c in self.selected_channels:
            self.traces[c].center_ampl()
        self.set_amplitudes()


    def set_frequencies(self, f0=None, f1=None):
        self.setting = True
        for c in self.selected_channels:
            if not f0 is None:
                self.specs[c].f0 = f0
            if not f1 is None:
                self.specs[c].f1 = f1
            if self.isVisible():
                for ax in self.axfys[c]:
                    ax.setYRange(self.specs[c].f0, self.specs[c].f1)
                for ax in self.axfxs[c]:
                    ax.setXRange(self.specs[c].f0, self.specs[c].f1)
        self.setting = False

            
    def update_frequencies(self, viewbox, frange):
        if self.setting:
            return
        self.set_frequencies(frange[0], frange[1])
        self.sigFrequenciesChanged.emit(frange[0], frange[1])
        
                
    def zoom_freq_in(self):
        for c in self.selected_channels:
            self.specs[c].zoom_freq_in()
        self.set_frequencies()
            
        
    def zoom_freq_out(self):
        for c in self.selected_channels:
            self.specs[c].zoom_freq_out()
        self.set_frequencies()
                
        
    def freq_down(self):
        for c in self.selected_channels:
            self.specs[c].freq_down()
        self.set_frequencies()

            
    def freq_up(self):
        for c in self.selected_channels:
            self.specs[c].freq_up()
        self.set_frequencies()


    def freq_home(self):
        for c in self.selected_channels:
            self.specs[c].freq_home()
        self.set_frequencies()

            
    def freq_end(self):
        for c in self.selected_channels:
            self.specs[c].freq_end()
        self.set_frequencies()


    def set_resolution(self, nfft=None, step_frac=None, dispatch=True):
        if self.setting:
            return
        self.setting = True
        if not isinstance(nfft, list):
            nfft = [nfft] * (np.max(self.selected_channels) + 1)
        if not isinstance(step_frac, list):
            step_frac = [step_frac] * (np.max(self.selected_channels) + 1)
        for c in self.selected_channels:
            self.specs[c].set_resolution(nfft[c], step_frac[c],
                                         self.isVisible())
        self.nfftw.setCurrentText(f'{self.specs[self.current_channel].nfft}')
        self.setting = False
        if dispatch:
            self.sigResolutionChanged.emit()

        
    def freq_resolution_down(self):
        for c in self.selected_channels:
            self.specs[c].freq_resolution_down()
        self.set_resolution()

        
    def freq_resolution_up(self):
        for c in self.selected_channels:
            self.specs[c].freq_resolution_up()
        self.set_resolution()


    def step_frac_down(self):
        for c in self.selected_channels:
            self.specs[c].step_frac_down()
        self.set_resolution()


    def step_frac_up(self):
        for c in self.selected_channels:
            self.specs[c].step_frac_up()
        self.set_resolution()

        
    def set_color_map(self, color_map=None, dispatch=True):
        if color_map is not None:
            self.color_map = color_map
        for cb in self.cbars:
            cb.setColorMap(self.color_maps[self.color_map])
        if dispatch:
            self.sigColorMapChanged.emit()

            
    def color_map_cycler(self):
        self.color_map += 1
        if self.color_map >= len(self.color_maps):
            self.color_map = 0
        self.set_color_map()


    def set_power(self, zmin=None, zmax=None, dispatch=True):
        self.setting = True
        if not isinstance(zmin, list):
            zmin = [zmin] * (np.max(self.selected_channels) + 1)
        if not isinstance(zmax, list):
            zmax = [zmax] * (np.max(self.selected_channels) + 1)
        for c in self.selected_channels:
            self.specs[c].set_power(zmin[c], zmax[c])
        self.setting = False
        if dispatch:
            self.sigPowerChanged.emit()


    def update_power(self, cbar):
        if self.setting:
            return
        self.set_power(cbar.levels()[0], cbar.levels()[1])


    def power_up(self):
        for c in self.selected_channels:
            self.specs[c].zmax += 5.0
            self.specs[c].zmin += 5.0
        self.set_power()


    def power_down(self):
        for c in self.selected_channels:
            self.specs[c].zmax -= 5.0
            self.specs[c].zmin -= 5.0
        self.set_power()


    def max_power_up(self):
        for c in self.selected_channels:
            self.specs[c].zmax += 5.0
        self.set_power()


    def max_power_down(self):
        for c in self.selected_channels:
            self.specs[c].zmax -= 5.0
        self.set_power()


    def min_power_up(self):
        for c in self.selected_channels:
            self.specs[c].zmin += 5.0
        self.set_power()


    def min_power_down(self):
        for c in self.selected_channels:
            self.specs[c].zmin -= 5.0
        self.set_power()


    def highpass_cutoff_up(self):
        highpass_cutoff = self.traces[self.current_channel].highpass_cutoff
        lowpass_cutoff = self.traces[self.current_channel].lowpass_cutoff
        step = 1.0
        if highpass_cutoff >= 1.0:
            step = 0.5*10**(floor(log10(highpass_cutoff)))
        elif highpass_cutoff == 0.0:
            step = 100.0
        highpass_cutoff += step
        if highpass_cutoff + step > lowpass_cutoff:
            highpass_cutoff = lowpass_cutoff - step
        if highpass_cutoff < 0:
            highpass_cutoff = 0
        self.update_filter(self.current_channel, highpass_cutoff,
                           None, True)


    def highpass_cutoff_down(self):
        highpass_cutoff = self.traces[self.current_channel].highpass_cutoff
        step = 1.0
        if highpass_cutoff >= 1.0:
            step = 0.5*10**(floor(log10(highpass_cutoff)))
            step = 0.5*10**(floor(log10(highpass_cutoff - 0.1*step)))
        highpass_cutoff -= step
        if highpass_cutoff < 0:
            highpass_cutoff = 0
        self.update_filter(self.current_channel, highpass_cutoff,
                           None, True)


    def lowpass_cutoff_up(self):
        lowpass_cutoff = self.traces[self.current_channel].lowpass_cutoff
        step = 1.0
        if lowpass_cutoff >= 1.0:
            step = 0.5*10**(floor(log10(lowpass_cutoff)))
        lowpass_cutoff += step
        if lowpass_cutoff > self.rate/2:
            lowpass_cutoff = self.rate/2
        self.update_filter(self.current_channel, None,
                           lowpass_cutoff, True)


    def lowpass_cutoff_down(self):
        highpass_cutoff = self.traces[self.current_channel].highpass_cutoff
        lowpass_cutoff = self.traces[self.current_channel].lowpass_cutoff
        step = 1.0
        if lowpass_cutoff >= 1.0:
            step = 0.5*10**(floor(log10(lowpass_cutoff)))
            step = 0.5*10**(floor(log10(lowpass_cutoff - 0.1*step)))
        if lowpass_cutoff < highpass_cutoff + step:
            return
        lowpass_cutoff -= step
        if lowpass_cutoff < highpass_cutoff + step:
            lowpass_cutoff = highpass_cutoff + step
        self.update_filter(self.current_channel, None,
                           lowpass_cutoff, True)


    def set_filter(self, highpass_cutoffs, lowpass_cutoffs):
        self.setting = True
        for c in range(self.data.channels):
            cf = c if c < len(highpass_cutoffs) else -1
            self.traces[c].set_filter(highpass_cutoffs[cf],
                                      lowpass_cutoffs[cf])
            self.axspecs[c].set_filter(highpass_cutoffs[cf],
                                       lowpass_cutoffs[cf])
        self.setting = False


    def update_filter(self, channel, highpass_cutoff,
                      lowpass_cutoff, set_spec=False):
        if channel in self.selected_channels:
            for c in self.selected_channels:
                self.traces[c].set_filter(highpass_cutoff, lowpass_cutoff)
                if c != channel or set_spec:
                    self.axspecs[c].set_filter(highpass_cutoff, lowpass_cutoff)
        else:
            self.traces[channel].set_filter(highpass_cutoff, lowpass_cutoff)
        self.sigFilterChanged.emit()


    def init_filter(self, highpass_cutoff, lowpass_cutoff):
        if highpass_cutoff is None:
            highpass_cutoff = 0.0
        if lowpass_cutoff is None:
            lowpass_cutoff = self.rate/2
        for t in self.traces:
            t.set_filter(highpass_cutoff, lowpass_cutoff)
        for axs in self.axspecs:
            axs.set_filter(highpass_cutoff, lowpass_cutoff)


    def add_to_show_channels(self, channels):
        if isinstance(channels, int):
            channels = [channels]
        for channel in channels:
            if not channel in self.show_channels:
                self.show_channels.append(channel)
        self.show_channels.sort()


    def add_to_selected_channels(self, channels):
        if isinstance(channels, int):
            channels = [channels]
        for channel in channels:
            if not channel in self.selected_channels:
                self.selected_channels.append(channel)
        self.selected_channels.sort()

        
    def all_channels(self):
        if self.selected_channels == self.show_channels:
            self.selected_channels = list(range(self.data.channels))
        else:
            self.selected_channels = list(self.show_channels)
        self.update_borders()


    def next_channel(self):
        idx = self.show_channels.index(self.current_channel)
        if idx + 1 < len(self.show_channels):
            self.current_channel = self.show_channels[idx + 1]
            self.selected_channels = [self.current_channel]
            self.update_borders()
        else:
            if self.show_channels[-1] < self.data.channels - 1:
                n = len(self.show_channels)
                if n > 1:
                    n -= 1
                if self.show_channels[-1] + n >= self.data.channels:
                    n = self.data.channels - 1 - self.show_channels[-1]
                self.add_to_show_channels(list(range(self.show_channels[-1] + 1,
                                                     self.show_channels[-1] + 1 + n)))
                del self.show_channels[:n]
                self.current_channel += 1
            self.selected_channels = [self.current_channel]
            self.set_channels()


    def previous_channel(self):
        idx = self.show_channels.index(self.current_channel)
        if idx > 0:
            self.current_channel = self.show_channels[idx - 1]
            self.selected_channels = [self.current_channel]
            self.update_borders()
        else:
            if self.show_channels[0] > 0:
                n = len(self.show_channels)
                if n > 1:
                    n -= 1
                if self.show_channels[0] < n:
                    n = self.show_channels[0]
                self.add_to_show_channels(list(range(self.show_channels[0] - n,
                                                 self.show_channels[0])))
                del self.show_channels[-n:]
                self.current_channel -= 1
            self.selected_channels = [self.current_channel]
            self.set_channels()


    def select_next_channel(self):
        show_selected_channels = [c for c in range(self.data.channels) if c in self.show_channels and c in self.selected_channels]
        if len(show_selected_channels) > 0:
            self.current_channel = show_selected_channels[-1]
        idx = self.show_channels.index(self.current_channel)
        if idx + 1 < len(self.show_channels):
            self.current_channel = self.show_channels[idx + 1]
            self.add_to_selected_channels(self.current_channel)
            self.update_borders()
        else:
            if self.show_channels[-1] < self.data.channels - 1:
                n = len(self.show_channels)
                if self.show_channels[-1] + n >= self.data.channels:
                    n = self.data.channels - 1 - self.show_channels[-1]
                self.add_to_show_channels(list(range(self.show_channels[-1] + 1,
                                                     self.show_channels[-1] + 1 + n)))
                del self.show_channels[:n]
            if self.current_channel < self.data.channels - 1:
                self.current_channel += 1
                self.add_to_selected_channels(self.current_channel)
            self.set_channels()


    def select_previous_channel(self):
        show_selected_channels = [c for c in range(self.data.channels) if c in self.show_channels and c in self.selected_channels]
        if len(show_selected_channels) > 0:
            self.current_channel = show_selected_channels[0]
        idx = self.show_channels.index(self.current_channel)
        if idx > 0:
            self.current_channel = self.show_channels[idx - 1]
            self.add_to_selected_channels(self.current_channel)
            self.update_borders()
        else:
            if self.show_channels[0] > 0:
                n = len(self.show_channels)
                if self.show_channels[0] < n:
                    n = self.show_channels[0]
                self.add_to_show_channels(list(range(self.show_channels[0] - n,
                                                     self.show_channels[0])))
                del self.show_channels[-n:]
            if self.current_channel > 0:
                self.current_channel -= 1
                self.add_to_selected_channels(self.current_channel)
            self.set_channels()

            
    def set_channels(self, show_channels=None, selected_channels=None,
                     current_channel=None):
        if self.setting:
            return
        self.setting = True
        if show_channels is not None:
            if self.data is None:
                self.channels = show_channels
                self.setting = False
                return
            self.show_channels = [c for c in show_channels if c < self.data.channels]
        if selected_channels is not None:
            self.selected_channels = [c for c in selected_channels if c < self.data.channels]
        if current_channel is not None:
            self.current_channel = current_channel
        # current channel must be in shown and selected channels:
        show_selected_channels = [c for c in range(self.data.channels) if c in self.show_channels and c in self.selected_channels]
        if not self.current_channel in show_selected_channels:
            for c in show_selected_channels:
                if c >= self.current_channel:
                    self.current_channel = c
                    break
            if not self.current_channel in show_selected_channels:
                self.current_channel = show_selected_channels[-1]
        for c in range(self.data.channels):
            self.figs[c].setVisible(c in self.show_channels)
            self.show_xticks(c, c == self.show_channels[-1])
            self.acts.channels[c].setChecked(c in self.show_channels)
        self.adjust_layout(self.width(), self.height())
        self.update_borders()
        self.setting = False
            
        
    def toggle_channel(self, channel):
        if self.setting:
            return
        if channel < 0 or channel >= self.data.channels:
            return
        if self.acts.channels[channel].isChecked():
            self.add_to_show_channels(channel)
            self.add_to_selected_channels(channel)
            self.set_channels()
        else:
            if channel in self.show_channels:
                self.show_channels.remove(channel)
                if len(self.show_channels) == 0:
                    c = channel + 1
                    if c >= self.data.channels:
                        c = 0
                    self.show_channels = [c]
                    self.add_to_selected_channels(c)
                if channel in self.selected_channels:
                    self.selected_channels.remove(channel)
                    if len(self.selected_channels) == 0:
                        for c in self.show_channels:
                            if c < channel:
                                self.current_channel = c
                            else:
                                break
                        self.selected_channels = [self.current_channel]
                #if len(self.show_channels) == 1:
                #    self.acts.channels[self.show_channels[0]].setCheckable(False)
                self.set_channels()
        self.setFocus()

        
    def show_channel(self, channel):
        if channel < 0 or channel >= self.data.channels:
            return
        if self.current_channel == channel and \
           self.show_channels == [channel]:
            self.set_channels(list(range(self.data.channels)))
        else:
            self.current_channel = channel
            self.add_to_selected_channels(channel)
            self.set_channels([channel])

        
    def hide_deselected_channels(self):
        show_channels = [c for c in self.show_channels if c in self.selected_channels]
        if len(show_channels) == 0:
            show_channels = [self.show_channels[0]]
        self.set_channels(show_channels)
        
        
    def set_panels(self, traces=None, specs=None, cbars=None, fulldata=None):
        if not traces is None:
            self.show_traces = traces
        if not specs is None:
            self.show_specs = specs
        if not cbars is None:
            self.show_cbars = cbars
        if not fulldata is None:
            self.show_fulldata = fulldata
        for axt, axs, cb in zip(self.axtraces, self.axspecs, self.cbars):
            axt.setVisible(self.show_traces)
            axs.setVisible(self.show_specs > 0)
            cb.setVisible(self.show_specs > 0 and self.show_cbars)
            if axt is self.axtraces[self.show_channels[-1]]:
                axs.getAxis('bottom').showLabel(not self.show_traces)
                axs.getAxis('bottom').setStyle(showValues=not self.show_traces)
                axt.getAxis('bottom').showLabel(self.show_traces)
                axt.getAxis('bottom').setStyle(showValues=self.show_traces)
        if self.datafig is not None:
            self.datafig.setVisible(self.show_fulldata)
        self.adjust_layout(self.width(), self.height())
            

    def toggle_traces(self):
        self.show_traces = not self.show_traces
        if not self.show_traces:
            self.show_specs = 1
        self.set_panels()
            

    def toggle_spectrograms(self):
        self.show_specs += 1
        if self.show_specs > 4:
            self.show_specs = 0
        if self.show_specs == 0:
            self.show_traces = True
        self.set_panels()

                
    def toggle_colorbars(self):
        self.show_cbars = not self.show_cbars
        self.set_panels()
            
                
    def toggle_fulldata(self):
        self.show_fulldata = not self.show_fulldata
        self.set_panels()
            
            
    def toggle_grids(self):
        self.grids -= 1
        if self.grids < 0:
            self.grids = 3
        for c in self.selected_channels:
            for ax in self.axgs[c]:
                ax.showGrid(x=(self.grids & 1) > 0, y=(self.grids & 2) > 0,
                            alpha=0.8)
                # fix grid bug:
                ax.getAxis('bottom').setGrid(False)
                ax.getAxis('left').setGrid(False)
                for axis in ['right', 'top']:
                    ax.showAxis(axis)
                    ax.getAxis(axis).setStyle(showValues=False)


    def set_zoom_mode(self, mode):
        for axs in self.axs:
            for ax in axs:
                ax.getViewBox().setMouseMode(mode)


    def zoom_back(self):
        for axs in self.axs:
            for ax in axs:
                ax.getViewBox().zoom_back()


    def zoom_forward(self):
        for axs in self.axs:
            for ax in axs:
                ax.getViewBox().zoom_forward()


    def zoom_home(self):
        for axs in self.axs:
            for ax in axs:
                ax.getViewBox().zoom_home()


    def set_region_mode(self, mode):
        self.region_mode = mode


    def region_menu(self, channel, vbox, rect):
        if self.region_mode == DataBrowser.zoom_region:
            vbox.zoom_region(rect)
        elif self.region_mode == DataBrowser.play_region:
            self.play_region(rect.left(), rect.right())
        elif self.region_mode == DataBrowser.save_region:
            self.save_region(rect.left(), rect.right())
        elif self.region_mode == DataBrowser.ask_region:
            menu = QMenu(self)
            zoom_act = menu.addAction('&Zoom')
            #analyze_act = menu.addAction('&Analyze')
            play_act = menu.addAction('&Play')
            save_act = menu.addAction('&Save as')
            #crop_act = menu.addAction('&Crop')
            act = menu.exec(QCursor.pos())
            if act is zoom_act:
                vbox.zoom_region(rect)
            elif act is play_act:
                self.play_region(rect.left(), rect.right())
            elif act is save_act:
                self.save_region(rect.left(), rect.right())
        vbox.hide_region()


    def play_scroll(self):
        if self.scroll_timer.isActive():
            self.scroll_timer.stop()
            self.scroll_step /= 2
        elif self.audio_timer.isActive():
            self.audio.stop()
            self.audio_timer.stop()
            for amarkers in self.audio_markers:
                for vmarker in amarkers:
                    vmarker.setValue(-1)
        else:
            self.play_window()
        

    def auto_scroll(self):
        if self.scroll_step == 0:
            self.scroll_step = 0.005
        elif self.scroll_step > 1.0:
            if self.scroll_timer.isActive():
                self.scroll_timer.stop()
            self.scroll_step = 0
            return
        else:
            self.scroll_step *= 2
        if not self.scroll_timer.isActive():
            self.scroll_timer.start(50)

        
    def scroll_further(self):
        if self.toffset + self.twindow > self.tmax:
            self.scroll_timer.stop()
            self.scroll_step /= 2
        else:
            self.set_times(self.toffset + self.twindow*self.scroll_step)


    def set_audio(self, rate_fac=None,
                  use_heterodyne=None, heterodyne_freq=None,
                  dispatch=True):
        if rate_fac is not None:
            self.audio_rate_fac = rate_fac
            if not dispatch:
                self.audiofacw.setCurrentText(f'{self.audio_rate_fac:g}')
        if use_heterodyne is not None:
            self.audio_use_heterodyne = use_heterodyne
        if heterodyne_freq is not None:
            self.audio_heterodyne_freq = float(heterodyne_freq)
            if not dispatch:
                self.audiohetfw.setValue(int(self.audio_heterodyne_freq/1000))
        if dispatch:
            self.sigAudioChanged.emit(self.audio_rate_fac,
                                      self.audio_use_heterodyne,
                                      self.audio_heterodyne_freq)


    def play_region(self, t0, t1):
        rate = self.rate
        i0 = int(np.round(t0*rate))
        i1 = int(np.round(t1*rate))
        if i1 > len(self.data):
            i1 = len(self.data)
            t1 = i1/rate
        n2 = (len(self.selected_channels)+1)//2
        playdata = np.zeros((i1-i0, min(2, len(self.selected_channels))))
        playdata[:,0] = np.mean(self.data[i0:i1, self.selected_channels[:n2]], 1)
        if len(self.selected_channels) > 1:
            playdata[:,1] = np.mean(self.data[i0:i1, self.selected_channels[n2:]], 1)
        if self.audio_use_heterodyne:
            # multiply with heterodyne frequency:
            heterodyne = np.sin(2*np.pi*self.audio_heterodyne_freq*np.arange(len(playdata))/rate)
            playdata = (playdata.T * heterodyne).T
            # low-pass filter and downsample:
            fcutoff = 20000.0
            sos = butter(2, 20000, 'low', output='sos', fs=rate)
            nstep = int(np.round(self.rate/(2*fcutoff)))
            if nstep < 1:
                nstep = 1
            playdata = sosfiltfilt(sos, playdata, 0)[::nstep]
            rate /= nstep
        fade(playdata, rate/self.audio_rate_fac, 0.1)
        self.audio.play(playdata, rate/self.audio_rate_fac, blocking=False)
        self.audio_time = t0
        self.audio_tmax = t1
        self.audio_timer.start(50)
        for c in range(self.data.channels):
            atime = self.audio_time if c in self.selected_channels else -1
            for vmarker in self.audio_markers[c]:
                vmarker.setValue(atime)


    def play_window(self):
        axt = None
        for axs in self.axts:
            if len(axs) > 0:
                axt = axs[0]
        if axt is None:
            return
        rect = axt.getViewBox().viewRect()
        self.play_region(rect.left(), rect.right())

        
    def mark_audio(self):
        self.audio_time += 0.05 / self.audio_rate_fac
        for amarkers in self.audio_markers:
            for vmarker in amarkers:
                if vmarker.value() >= 0:
                    vmarker.setValue(self.audio_time)
        if self.audio_time > self.audio_tmax:
            self.audio_timer.stop()
            for amarkers in self.audio_markers:
                for vmarker in amarkers:
                    vmarker.setValue(-1)

                    
    def save_region(self, t0, t1):

        def secs_to_str(time):
            hours = time//3600
            time -= 3600*hours
            mins = time//60
            time -= 60*mins
            secs = int(np.floor(time))
            time -= secs
            msecs = f'{1000*time:03.0f}ms'
            if hours > 0:
                return f'{hours}h{mins}m{secs}s{msecs}'
            elif mins > 0:
                return f'{mins}m{secs}s{msecs}'
            elif secs > 0:
                return f'{secs}s{msecs}'
            else:
                return msecs

        i0 = int(np.round(t0*self.rate))
        i1 = int(np.round(t1*self.rate))
        name = os.path.splitext(os.path.basename(self.file_path))[0]
        #if self.channel > 0:
        #    filename = f'{name}-{channel:d}-{t0:.4g}s-{t1s:.4g}s.wav'
        t0s = secs_to_str(t0)
        t1s = secs_to_str(t1)
        file_name = f'{name}-{t0s}-{t1s}.wav'
        formats = available_formats()
        for f in ['MP3', 'OGG', 'WAV']:
            if 'WAV' in formats:
                formats.remove(f)
                formats.insert(0, f)
        filters = ['All files (*)'] + [f'{f} files (*.{f}, *.{f.lower()})' for f in formats]
        file_path = os.path.join(os.path.dirname(self.file_path), file_name)
        file_path = QFileDialog.getSaveFileName(self, 'Save region as',
                                                file_path,
                                                ';;'.join(filters))[0]
        if file_path:
            md = deepcopy(self.data.metadata())
            update_starttime(md, t0, self.rate)
            hkey = 'CodingHistory'
            if 'BEXT' in md:
                hkey = 'BEXT.' + hkey
            bext_code = bext_history_str(self.data.encoding,
                                         self.rate, self.data.channels)
            add_history(md, bext_code + f',T=cut out {t0s}-{t1s}: {os.path.basename(file_path)}', hkey, bext_code + f',T={self.file_path}')
            locs, labels = self.marker_data.get_markers(self.rate)
            sel = (locs[:,0] + locs[:,1] >= i0) & (locs[:,0] <= i1)
            locs = locs[sel]
            labels = labels[sel]
            try:
                write_data(file_path, self.data[i0:i1,self.selected_channels],
                           self.rate, self.data.ampl_max, self.data.unit,
                           md, locs, labels, encoding=self.data.encoding)
                print(f'saved region to "{os.path.relpath(file_path)}"')
            except PermissionError as e:
                print(f'failed to save region to "{os.path.relpath(file_path)}": permission denied')

        
    def save_window(self):
        self.save_region(self.toffset, self.toffset + self.twindow)



try:
    from PyQt5.QtCore import Signal
except ImportError:
    from PyQt5.QtCore import pyqtSignal as Signal
import pyqtgraph as pg
from .selectviewbox import SelectViewBox
from .timeaxisitem import TimeAxisItem
from .yaxisitem import YAxisItem


class SpectrumPlot(pg.PlotItem):

    
    sigSelectedRegion = Signal(object, object, object)
    sigUpdateFilter = Signal(object, object, object)


    def __init__(self, channel, xwidth, starttime, fmax):

        self.channel = channel

        # view box:
        view = SelectViewBox(channel)
        
        # axis:
        bottom_axis = TimeAxisItem(orientation='bottom', showValues=True)
        bottom_axis.setLabel('Time', 's', color='black')
        bottom_axis.showLabel(False)
        bottom_axis.setStyle(showValues=False)
        bottom_axis.setPen('white')
        bottom_axis.setTextPen('black')
        bottom_axis.setStartTime(starttime)
        top_axis = TimeAxisItem(orientation='top', showValues=False)
        top_axis.setStartTime(starttime)
        left_axis = YAxisItem(orientation='left', showValues=True)
        left_axis.setLabel('Frequency', 'Hz', color='black')
        left_axis.setPen('white')
        left_axis.setTextPen('black')
        left_axis.setWidth(8*xwidth)
        right_axis = YAxisItem(orientation='right', showValues=False)

        # plot:
        pg.PlotItem.__init__(self,  viewBox=view,
                             axisItems={'bottom': bottom_axis,
                                        'top': top_axis,
                                        'left': left_axis,
                                        'right': right_axis})

        # design:
        self.getViewBox().setBackgroundColor('black')
        self.getViewBox().setDefaultPadding(padding=0.0)

        # ranges:
        self.setLimits(xMin=0, yMin=0.0, yMax=fmax,
                       minYRange=0.1, maxYRange=fmax)

        # functionality:
        self.enableAutoRange(False, False)
        self.hideButtons()
        self.setMenuEnabled(False)

        # filter handles:
        self.highpass_cutoff = 0
        self.lowpass_cutoff = 0
        self.highpass_handle = pg.InfiniteLine(angle=0, movable=True)
        self.highpass_handle.setPen(pg.mkPen('white', width=2))
        self.highpass_handle.addMarker('o', position=0.75, size=6)
        self.highpass_handle.setZValue(100)
        self.highpass_handle.setBounds((0, fmax))
        self.highpass_handle.setValue(self.highpass_cutoff)
        self.highpass_handle.sigPositionChangeFinished.connect(self.highpass_changed)
        self.addItem(self.highpass_handle, ignoreBounds=True)
        self.lowpass_handle = pg.InfiniteLine(angle=0, movable=True)
        self.lowpass_handle.setPen(pg.mkPen('white', width=2))
        self.lowpass_handle.addMarker('o', position=0.75, size=6)
        self.lowpass_handle.setZValue(100)
        self.lowpass_handle.setBounds((0, fmax))
        self.lowpass_handle.setValue(self.lowpass_cutoff)
        self.lowpass_handle.sigPositionChangeFinished.connect(self.lowpass_changed)
        self.addItem(self.lowpass_handle, ignoreBounds=True)

        # audio marker:
        self.vmarker = pg.InfiniteLine(angle=90, movable=False)
        self.vmarker.setPen(pg.mkPen('white', width=2))
        self.vmarker.setZValue(100)
        self.vmarker.setValue(-1)
        self.addItem(self.vmarker, ignoreBounds=True)

        # cross hair:
        self.xline = pg.InfiniteLine(angle=90, movable=False)
        self.xline.setPen(pg.mkPen('white', width=1))
        self.xline.setZValue(100)
        self.xline.setValue(0)
        self.xline.setVisible(False)
        self.addItem(self.xline, ignoreBounds=True)
        
        self.yline = pg.InfiniteLine(angle=0, movable=False)
        self.yline.setPen(pg.mkPen('white', width=1))
        self.yline.setZValue(100)
        self.yline.setValue(0)
        self.yline.setVisible(False)
        self.addItem(self.yline, ignoreBounds=True)

        # previous cross hair marker:
        self.prev_marker = pg.ScatterPlotItem(
            size=14,
            pen=pg.mkPen('white'),
            brush=pg.mkBrush((255, 255, 255, 128)),
            symbol='o',
            hoverable=False
        )
        self.prev_marker.setZValue(20)
        self.addItem(self.prev_marker, ignoreBounds=True)

        # signals:
        view.sigSelectedRegion.connect(self.sigSelectedRegion)


    def enableStartTime(self, enable):
        """ Enable addition of start time to tick labels.

        Parameters
        ----------
        enable: bool
            If True enable addition of start time to tick labels.
        """
        self.getAxis('bottom').enableStartTime(enable)
        self.getAxis('top').enableStartTime(enable)


    def set_filter(self, highpass_cutoff=None, lowpass_cutoff=None):
        if highpass_cutoff is not None:
            self.highpass_cutoff = highpass_cutoff
            self.highpass_handle.setValue(self.highpass_cutoff)
        if lowpass_cutoff is not None:
            self.lowpass_cutoff = lowpass_cutoff
            self.lowpass_handle.setValue(self.lowpass_cutoff)


    def highpass_changed(self):
        self.set_filter(highpass_cutoff=self.highpass_handle.value())
        self.sigUpdateFilter.emit(self.channel, self.highpass_cutoff,
                                  self.lowpass_cutoff)
        

    def lowpass_changed(self):
        self.set_filter(lowpass_cutoff=self.lowpass_handle.value())
        self.sigUpdateFilter.emit(self.channel, self.highpass_cutoff,
                                  self.lowpass_cutoff)




import sys
import os
import platform
import argparse
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.mlab as ml
import matplotlib.colors as mc
import matplotlib.widgets as widgets
import scipy.signal as sig
from audioio import load_audio, write_audio, PlayAudio, fade
from .version import __version__, __year__
from .configfile import ConfigFile

    
cfg = ConfigFile()

cfg.add_section('Plotting:')
cfg.add('maxpixel', 50000, '', 'Either maximum number of data points to be plotted or zero for plotting all data points.')

cfg.add_section('Envelope:')
cfg.add('envcutofffreq', 100.0, 'Hz', 'Cutoff frequency of the low-pass filter used for computing the envelope from the squared signal.')
cfg.add('envthreshfac', 0.2, '', 'Threshold for peak detection in envelope is this factor times the 95% - 5% percentile of the envelope.')

cfg.add_section('Power spectrum estimation:')
cfg.add('minPSDAverages', 3, '', 'Minimum number of fft averages for estimating the power spectrum.')

cfg.add_section('Thresholds for peak detection in power spectra:')
cfg.add('threshold', 10.0, 'dB', 'Threshold for all peaks.\n If set to 0.0, estimate threshold from histogram.')

cfg.add_section('Threshold estimation:\nIf no thresholds are specified, they are estimated from the histogram of the decibel power spectrum.')
cfg.add('noiseFactor', 12.0, '', 'Factor for multiplying std of noise floor for lower threshold.')

cfg.add_section('Items to display:')
cfg.add('displayHelp', False, '', 'Display help on key bindings')
cfg.add('labelFrequency', True, '', 'Display the frequency of the peak')
cfg.add('labelPower', True, '', 'Display the power of the peak')
cfg.add('labelWidth', True, '', 'Display the width of the peak')

cfg.add_section('Debugging:')
cfg.add('verboseLevel', 0, '', '0=off upto 4 very detailed')

    
###############################################################################
## filter and envelope:

def highpass_filter(data, rate, cutoff):
    sos = sig.butter(2, cutoff, 'highpass', fs=rate, output='sos')
    fdata = sig.sosfiltfilt(sos, data)
    return fdata


def bandpass_filter(data, rate, lowf=5500.0, highf=7500.0):
    """
    Bandpass filter the signal.
    """
    sos = sig.butter(2, (lowf, highf), 'bandpass', fs=rate, output='sos')
    fdata = sig.sosfiltfilt(sos, data)
    return fdata


def envelope(data, rate, freq=100.0):
    sos = sig.butter(2, freq, 'lowpass', fs=rate, output='sos')
    envelope = np.sqrt(2)*sig.sosfiltfilt(sos, np.abs(data))
    return envelope


###############################################################################
## peak detection:

def detect_peaks_fixed(data, threshold):
    """Detect peaks and troughs using a fixed, relative threshold.

    From https://github.com/bendalab/thunderlab/blob/master/thunderlab/eventdetection.py

    Parameters
    ----------
    data: array
        An 1-D array of input data where peaks are detected.
    threshold: float
        A positive number setting the detection threshold,
        i.e. the minimum distance between peaks and troughs.
    
    Returns
    -------
    peaks: array of ints
        An array of indices of detected peaks.
    troughs: array of ints
        An array of indices of detected troughs.
    """
    peaks = []
    troughs = []

    # initialize:
    direction = 0
    min_inx = 0
    max_inx = 0
    min_value = data[0]
    max_value = min_value

    # loop through the data:
    for index, value in enumerate(data):
        # rising?
        if direction > 0:
            if value > max_value:
                # update maximum element:
                max_inx = index
                max_value = value
            # otherwise, if the new value is falling below
            # the maximum value minus the threshold:
            # the maximum is a peak!
            elif value <= max_value - threshold:
                peaks.append(max_inx)
                # change direction:
                direction = -1
                # store minimum element:
                min_inx = index
                min_value = value

        # falling?
        elif direction < 0:
            if value < min_value:
                # update minimum element:
                min_inx = index
                min_value = value
            # otherwise, if the new value is rising above
            # the minimum value plus the threshold:
            # the minimum is a trough!
            elif value >= min_value + threshold:
                troughs.append(min_inx)
                # change direction:
                direction = +1
                # store maximum element:
                max_inx = index
                max_value = value

        # don't know direction yet:
        else:
            if value <= max_value - threshold:
                direction = -1  # falling
            elif value >= min_value + threshold:
                direction = 1  # rising
                
            if value > max_value:
                # update maximum element:
                max_inx = index
                max_value = value
            elif value < min_value:
                # update minimum element:
                min_inx = index
                min_value = value

    return np.asarray(peaks, dtype=np.int), \
           np.asarray(troughs, dtype=np.int)


def threshold_estimate(data, noise_factor):
    """
    Estimate noise standard deviation from histogram
    for usefull peak-detection thresholds.

    The standard deviation of the noise floor without peaks is estimated from
    the histogram of the data at 1/sqrt(e) relative height.

    Args:
        data: the data from which to estimate the thresholds
        noise_factor (float): multiplies the estimate of the standard deviation
                              of the noise to result in the threshold

    Returns:
        threshold (float): the threshold above the noise floor
        center: (float): estimate of the median of the data without peaks
    """
    
    # estimate noise standard deviation:
    # XXX what about the number of bins for small data sets?
    hist, bins = np.histogram(data, 100, density=True)
    inx = hist > np.max(hist) / np.sqrt(np.e)
    lower = bins[inx][0]
    upper = bins[inx][-1] # needs to return the next bin
    center = 0.5*(lower+upper)
    noisestd = 0.5*(upper-lower)
    
    # threshold:
    threshold = noise_factor*noisestd

    return threshold, center


def peak_size_width(time, data, peak_indices, trough_indices,
                    peak_frac=0.75, base='closest'):
    """Compute size and width of each peak.

    Parameters
    ----------
    time: array
        Time, must not be `None`.
    data: array
        The data with the peaks.
    peak_indices: array
        Indices of the peaks.
    trough_indices: array
        Indices of the troughs.
    peak_frac: float
        Fraction of peak height where its width is measured.
    base: string
        Height and width of peak is measured relative to
        - 'left': trough to the left
        - 'right': trough to the right
        - 'min': the minimum of the two troughs to the left and to the right
        - 'max': the maximum of the two troughs to the left and to the right
        - 'mean': mean of the throughs to the left and to the rigth
        - 'closest': trough that is closest to peak
    
    Returns
    -------
    peaks: 2-D array
        First dimension is the peak index. Second dimension is
        time, height (value of data at the peak),
        size (peak height minus height of closest trough),
        width (at `peak_frac` size), 0.0 (count) of the peak. See `peak_width()`.

    Raises
    ------
    ValueError:
        If an invalid value is passed to `base`.
    """
    def left_base(data, left_inx, right_inx, peak_inx):
        return data[left_inx] 
    def right_base(data, left_inx, right_inx, peak_inx):
        return data[right_inx] 
    def min_base(data, left_inx, right_inx, peak_inx):
        return min(data[left_inx], data[right_inx])
    def max_base(data, left_inx, right_inx, peak_inx):
        return max(data[left_inx], data[right_inx])
    def mean_base(data, left_inx, right_inx, peak_inx):
        return np.mean((data[left_inx], data[right_inx]))
    def closest_base(data, left_inx, right_inx, peak_inx):
        return data[left_inx] if peak_inx-left_inx <= right_inx-peak_inx else data[right_inx]
    
    peaks = np.zeros((len(peak_indices), 5))
    if len(peak_indices) == 0:
        return peaks
    # time point of peaks:
    peaks[:, 0] = time[peak_indices]
    # height of peaks:
    peaks[:, 1] = data[peak_indices]
    # we need a trough before and after each peak:
    peak_inx = np.asarray(peak_indices, dtype=int)
    trough_inx = np.asarray(trough_indices, dtype=int)
    if len(trough_inx) == 0 or peak_inx[0] < trough_inx[0]:
         trough_inx = np.hstack((0, trough_inx))

    if peak_inx[-1] > trough_inx[-1]:
         trough_inx = np.hstack((trough_inx, len(data)-1))
    # base for size of peaks:
    base_func = closest_base
    if base == 'left':
        base_func = left_base
    elif base == 'right':
        base_func = right_base
    elif base == 'min':
        base_func = min_base
    elif base == 'max':
        base_func = max_base
    elif base == 'mean':
        base_func = mean_base
    elif base == 'closest':
        base_func = closest_base

    else:
        raise ValueError('Invalid value for base (%s)' % base)
    # size and width of peaks:
    for j, pi in enumerate(peak_inx):
        li = trough_inx[j]
        ri = trough_inx[j+1]
        baseval = base_func(data, li, ri, pi)
        thresh = baseval*(1.0-peak_frac) + data[pi]*peak_frac
        inx = li + np.argmax(data[li:ri] > thresh)
        if inx > 0:
            ti0 = np.interp(thresh, data[inx-1:inx+1], time[inx-1:inx+1])
        else:
            ti0 = time[0]
        inx = ri - np.argmax(data[ri:li:-1] > thresh)
        if inx+1 < len(data):
            ti1 = np.interp(thresh, data[inx+1:inx-1:-1], time[inx+1:inx-1:-1])
        else:
            ti1 = time[-1]
        if np.any(np.isfinite((data[pi], baseval))):
            peaks[j, 2] = data[pi] - baseval
        peaks[j, 3] = ti1 - ti0
    return peaks


def psd_peaks(psd_freqs, psd, cfg):
    """
    Detect peaks in power spectrum.

    Args:
        psd_freqs (array): frequencies of the power spectrum
        psd (array): power spectrum
        cfg (dict): configuration parameter

    Returns:
        all_freqs (2-d array): peaks in the power spectrum
                  detected with low threshold
                  [frequency, power, size, width, double use count]
        threshold (float): the relative threshold for detecting all peaks in the decibel spectrum
        center (float): the baseline level of the power spectrum
    """
    
    verbose = cfg.value('verboseLevel')

    if verbose > 0:
        print()
        print(70*'#')
        print('##### psd_peaks', 48*'#')
    
    # decibel power spectrum:
    log_psd = 10.0*np.log10(psd)

    # thresholds:
    threshold = cfg.value('threshold')
    center = 0.0
    if cfg.value('threshold') <= 0.0:
        n = len(log_psd)
        threshold, center = threshold_estimate(log_psd[2*n/3:n*9/10],
                                                cfg.value('noiseFactor'))
        if verbose > 1:
            print()
            print('threshold=', threshold, center+threshold)
            print('center=', center)
    
    # detect peaks in decibel power spectrum:
    peaks, troughs = detect_peaks_fixed(log_psd, threshold)
    all_peaks = peak_size_width(psd_freqs, log_psd, peaks, troughs)
    
    return all_peaks, threshold, center

    
###############################################################################
## plotting etc.
    
class SignalPlot:
    def __init__(self, data, samplingrate, unit, filename, channel, path):
        self.filepath = ''
        if platform.system() == 'Windows':
            self.filepath = path
        self.filename = filename
        self.channel = channel
        self.rate = samplingrate
        self.data = data
        self.unit = unit
        self.envcutofffreq = cfg.value('envcutofffreq')
        self.envcutoff_artist = None
        self.envthreshfac = cfg.value('envthreshfac')
        self.envelope = envelope(self.data, self.rate, self.envcutofffreq)
        self.envpower = None
        self.envfreqs = None
        self.time = np.arange(0.0, len(self.data))/self.rate
        self.toffset = 0.0
        self.twindow = 8.0
        if self.twindow > self.time[-1]:
            self.twindow = np.round(2**(np.floor(np.log(self.time[-1]) / np.log(2.0)) + 1.0))
        self.ymin = -1.0
        self.ymax = +1.0
        self.trace_artist = None
        self.envelope_artist = None
        self.show_envelope = True
        self.spectrogram_artist = None
        self.fmin = 0.0
        self.fmax = 0.0
        self.decibel = True
        self.fresolution = 300.0
        self.power_label = None
        self.all_peaks_artist = None
        self.power_artist = None
        self.power_frequency_label = None
        self.envpower_label = None
        self.envpower_artist = None
        self.envpower_peaks_artist = None
        self.envpeaks_artist = None
        self.help = cfg.value('displayHelp')
        self.helptext = []
        self.allpeaks = []
        self.peak_specmarker = []
        self.peak_annotation = []
        self.env_peak_annotation = []
        self.analysis_file = None
        self.audio = PlayAudio()

        # set key bindings:
        plt.rcParams['keymap.fullscreen'] = 'ctrl+f'
        plt.rcParams['keymap.pan'] = 'ctrl+m'
        plt.rcParams['keymap.quit'] = 'ctrl+w, alt+q, q'
        plt.rcParams['keymap.yscale'] = ''
        plt.rcParams['keymap.xscale'] = ''
        plt.rcParams['keymap.grid'] = ''
        plt.rcParams['keymap.save'] = ''
        if 'keymap.all_axes' in plt.rcParams:
            plt.rcParams['keymap.all_axes'] = ''
        
        # the figure:
        plt.ioff()
        self.fig = plt.figure(figsize=(15, 9), constrained_layout=True)
        self.fig.canvas.manager.set_window_title('AUDIoANalyzer: ' + self.filename + ' channel {0:d}'.format(self.channel))
        self.fig.canvas.mpl_connect('key_press_event', self.keypress)
        self.fig.canvas.mpl_connect('button_press_event', self.buttonpress)
        self.fig.canvas.mpl_connect('pick_event', self.onpick)
        gs = self.fig.add_gridspec(2, 1, height_ratios=[5, 2])
        gst = gs[0].subgridspec(2, 1, hspace=0.0)
        gsp = gs[1].subgridspec(1, 2)
        # trace plot:
        self.axt = self.fig.add_subplot(gst[0])
        self.axt.xaxis.set_major_locator(plt.NullLocator())
        self.axt.set_ylabel('Amplitude [{:s}]'.format(self.unit))
        self.span = widgets.SpanSelector(self.axt, self.analyse_trace,
                                         direction='horizontal')
        ht = self.axt.text(0.98, 0.05, '(ctrl+) page and arrow up, down, home, end: scroll', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.15, '+, -, X, x: zoom in/out', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.25, 'y,Y,v,V: zoom amplitudes', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.35, 'p,P: play audio (display,all)', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.45, 'ctrl-f: full screen', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.55, 'w, W: plot waveform/spectrum into png file', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.65, 's: save current segment to wav file', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.75, 'q: quit', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        ht = self.axt.text(0.98, 0.85, 'h: toggle this help', ha='right', transform=self.axt.transAxes)
        self.helptext.append(ht)
        #self.axt.set_xticklabels([])
        # spectrogram:
        self.axs = self.fig.add_subplot(gst[1])
        #self.axs.sharex(self.axt)
        self.axs.get_shared_x_axes().join(self.axs, self.axt)
        self.axs.set_xlabel('Time [seconds]')
        self.axs.set_ylabel('Frequency [Hz]')
        # power spectrum:
        self.axp = self.fig.add_subplot(gsp[0])
        ht = self.axp.text(0.98, 0.9, 'r, R: frequency resolution', ha='right', transform=self.axp.transAxes)
        self.helptext.append(ht)
        ht = self.axp.text(0.98, 0.8, 'f, F: zoom', ha='right', transform=self.axp.transAxes)
        self.helptext.append(ht)
        ht = self.axp.text(0.98, 0.7, '(ctrl+) left, right: move', ha='right', transform=self.axp.transAxes)
        self.helptext.append(ht)
        ht = self.axp.text(0.98, 0.6, 'd: toggle decibel', ha='right', transform=self.axp.transAxes)
        self.helptext.append(ht)
        ht = self.axp.text(0.98, 0.3, 'left mouse: show peak properties', ha='right', transform=self.axp.transAxes)
        self.helptext.append(ht)
        ht = self.axp.text(0.98, 0.2, 'shift/ctrl + left/right mouse: goto previous/next harmonic', ha='right', transform=self.axp.transAxes)
        self.helptext.append(ht)
        ht = self.axp.text(0.98, 0.1, 'S: save current spectrum to csv file', ha='right', transform=self.axp.transAxes)
        self.helptext.append(ht)
        # power spectrum of envelope:
        self.axpe = self.fig.add_subplot(gsp[1])
        self.envcutoff_artist = self.axpe.text(0.05, 0.1, 'cutoff={:.0f} Hz'.format(self.envcutofffreq), transform=self.axpe.transAxes)
        ht = self.axpe.text(0.98, 0.9, 'c, C: envelope cutoff frequency', ha='right', transform=self.axpe.transAxes)
        self.helptext.append(ht)
        ht = self.axpe.text(0.98, 0.8, 't, T: threshold for envelope peak detection', ha='right', transform=self.axpe.transAxes)
        self.helptext.append(ht)
        ht = self.axpe.text(0.98, 0.7, 'e: toggle envelope', ha='right', transform=self.axpe.transAxes)
        self.helptext.append(ht)
        ht = self.axpe.text(0.98, 0.6, 'E: save envelope and its spectrum to files', ha='right', transform=self.axpe.transAxes)
        self.helptext.append(ht)
        # plot:
        for ht in self.helptext:
            ht.set_visible(self.help)
        self.update_plots(False)
        plt.show()

    def __del__(self):
        if not self.analysis_file is None:
            self.analysis_file.close()
        if self.audio is not None:
            self.audio.close()

    def compute_psd(self, t0, t1):
        nfft = int(np.round(2**(np.floor(np.log(self.rate/self.fresolution) / np.log(2.0)) + 1.0)))
        if nfft < 16:
            self.fresolution *= 0.5
            nfft = 16
        t00 = t0
        t11 = t1
        w = t11-t00
        minw = int(nfft*(cfg.value('minPSDAverages')+1)//2)
        if t11-t00 < minw:
            w = minw
            t11 = t00 + w
        if t11 >= len(self.data):
            t11 = len(self.data)
            t00 = t11 - w
        if t00 < 0:
            t00 = 0
            t11 = w           
        power, freqs = ml.psd(self.data[t00:t11], NFFT=nfft, noverlap=nfft//2, Fs=self.rate, detrend=ml.detrend_mean)
        return power, freqs, nfft, w

    def remove_peak_annotation(self):
        for fm in self.peak_specmarker:
            fm.remove()
        self.peak_specmarker = []
        for fa in self.peak_annotation:
            fa.remove()
        self.peak_annotation = []

    def annotate_peak(self, peak):
        # marker:
        m, = self.axs.plot([self.toffset+0.01*self.twindow], [peak[0]], linestyle='None',
                            color='k', marker='o', ms=10.0, mec=None, mew=0.0, zorder=2)
        self.peak_specmarker.append(m)
        # annotation:
        fwidth = self.fmax - self.fmin
        pt = []
        if cfg.value('labelFrequency'):
            pt.append(r'$f=${:.1f} Hz'.format(peak[0]))
        if cfg.value('labelPower'):
            pt.append(r'$p=${:.1f} dB'.format(peak[1]))
        if cfg.value('labelWidth'):
            pt.append(r'$\Delta f=${:.0f} Hz'.format(peak[3]))
        ypeak = peak[1] if self.decibel else 10.0**(0.1*peak[1])
        self.peak_annotation.append(self.axp.annotate('\n'.join(pt), xy=(peak[0], ypeak),
                       xytext=(peak[0]+0.03*fwidth, ypeak),
                       bbox=dict(boxstyle='round',facecolor='white'),
                       arrowprops=dict(arrowstyle='-')))

    def remove_env_peak_annotation(self):
        for fa in self.env_peak_annotation:
            fa.remove()
        self.env_peak_annotation = []

    def annotate_env_peak(self, peak):
        fwidth = 100.0
        pt = []
        if cfg.value('labelFrequency'):
            pt.append(r'$f=${:.1f} Hz'.format(peak[0]))
        if cfg.value('labelPower'):
            pt.append(r'$p=${:.1f} dB'.format(peak[1]))
        if cfg.value('labelWidth'):
            pt.append(r'$\Delta f=${:.1f} Hz'.format(peak[3]))
        ypeak = peak[1] if self.decibel else 10.0**(0.1*peak[1])
        self.env_peak_annotation.append(self.axpe.annotate('\n'.join(pt), xy=(peak[0], ypeak),
                       xytext=(peak[0]+0.03*fwidth, ypeak),
                       bbox=dict(boxstyle='round',facecolor='white'),
                       arrowprops=dict(arrowstyle='-')))
            
    def update_plots(self, draw=True):
        self.remove_peak_annotation()
        self.remove_env_peak_annotation()
        # trace:
        self.axt.set_xlim(self.toffset, self.toffset+self.twindow)
        t0 = int(np.round(self.toffset*self.rate))
        t1 = int(np.round((self.toffset+self.twindow)*self.rate))
        tstep = 1
        if cfg.value('maxpixel') > 0:
            tstep = int((t1-t0)//cfg.value('maxpixel'))
            if tstep < 1:
                tstep = 1
        if self.trace_artist is None:
            self.trace_artist, = self.axt.plot(self.time[t0:t1:tstep], self.data[t0:t1:tstep], color='tab:blue')
        else:
            self.trace_artist.set_data(self.time[t0:t1:tstep], self.data[t0:t1:tstep])
        if self.envelope_artist is None:
            self.envelope_artist,  = self.axt.plot(self.time[t0:t1:tstep], self.envelope[t0:t1:tstep], color='tab:red')
        else:
            self.envelope_artist.set_data(self.time[t0:t1:tstep], self.envelope[t0:t1:tstep])
        self.axt.set_ylim(self.ymin, self.ymax)

        # compute power spectrum:
        nfft = int(np.round(2**(np.floor(np.log(self.rate/self.fresolution) / np.log(2.0)) + 1.0)))
        if nfft < 16:
            self.fresolution *= 0.5
            nfft = 16
        t00 = t0
        t11 = t1
        w = t11-t00
        minw = int(nfft*(cfg.value('minPSDAverages')+1)//2)
        if t11-t00 < minw:
            w = minw
            t11 = t00 + w
        if t11 >= len(self.data):
            t11 = len(self.data)
            t00 = t11 - w
        if t00 < 0:
            t00 = 0
            t11 = w           
        self.power, self.freqs = ml.psd(self.data[t00:t11], NFFT=nfft, noverlap=nfft/2, Fs=self.rate, detrend=ml.detrend_mean)
        # detect peaks:
        self.allpeaks, lowth, center = psd_peaks(self.freqs, self.power, cfg)

        # spectrogram:
        t2 = t1 + nfft
        specpower, freqs, bins = ml.specgram(self.data[t0:t2], NFFT=nfft, Fs=self.rate, noverlap=nfft/2,
                                              detrend=ml.detrend_mean)
        specpower[specpower<=0.0] = np.min(specpower[specpower>0.0]) # remove zeros
        z = 10.*np.log10(specpower)
        z = np.flipud(z)
        sstep = z.shape[1]//2000
        if sstep < 1:
            sstep = 1
        extent = self.toffset, self.toffset+np.amax(bins), freqs[0], freqs[-1]
        self.axs.set_xlim(self.toffset, self.toffset+self.twindow)
        if self.spectrogram_artist is None:
            self.fmax = np.round((freqs[-1])/1000.0)*1000.0
            min = np.percentile(z, 70.0)
            max = np.percentile(z, 99.9) + 5.0
            cm = plt.get_cmap('jet')
            self.spectrogram_artist = self.axs.imshow(z[:,::sstep], aspect='auto',
                                                         extent=extent, vmin=min, vmax=max,
                                                         cmap=cm, zorder=1)
            #self.spectrogram_artist = self.axs.pcolormesh(bins, freqs, z,
            #                                                vmin=min, vmax=max,
            #                                                cmap=cm, zorder=1)
        else:
            self.spectrogram_artist.set_data(z[:,::sstep])
            self.spectrogram_artist.set_extent(extent)
        self.axs.set_ylim(self.fmin, self.fmax)

        # power spectrum:
        df = self.freqs[1]-self.freqs[0]
        if df >= 1000.0:
            dfs = '%.3gkHz' % (0.001*df)
        else:
            dfs = '%.3gHz' % df
        tw = float(w)/self.rate
        if tw < 1.0:
            tws = '%.3gms' % (1000.0*tw)
        else:
            tws = '%.3gs' % (tw)
        a = 2*w/nfft-1 # number of ffts
        if self.power_frequency_label is None:
            self.power_frequency_label = self.axp.set_xlabel(r'Frequency [Hz] (nfft={:d}, $\Delta f$={:s}: T={:s}/{:.0f})'.format(nfft, dfs, tws, a))
        else:
            self.power_frequency_label.set_text(r'Frequency [Hz] (nfft={:d}, $\Delta f$={:s}: T={:s}/{:.0f})'.format(nfft, dfs, tws, a))
        self.axp.set_xlim(self.fmin, self.fmax)
        if self.power_label is None:
            self.power_label = self.axp.set_ylabel('Signal power')
        if self.decibel:
            peak_power = self.allpeaks[:,1]
            self.power = 10.0*np.log10(self.power)
            pmin = np.min(self.power[self.freqs<self.fmax])
            pmin = np.floor(pmin/10.0)*10.0
            pmax = np.max(self.power[self.freqs<self.fmax])
            pmax = np.ceil(pmax/10.0)*10.0
            doty = pmax-5.0
            self.power_label.set_text('Signal power [dB]')
            self.axp.set_ylim(pmin, pmax)
        else:
            peak_power = 10.0**(0.1*self.allpeaks[:,1])
            pmax = np.max(self.power[self.freqs<self.fmax])
            doty = pmax
            pmax *= 1.1
            self.power_label.set_text('Signal power')
            self.axp.set_ylim(0.0, pmax)
        if self.all_peaks_artist is None:
            if len(self.allpeaks) > 0:
                self.all_peaks_artist, = self.axp.plot(self.allpeaks[:,0], peak_power,
                                                       'o', color='tab:red')
            else:
                self.all_peaks_artist, = self.axp.plot([], [], 'o', color='tab:red')
        else:
            if len(self.allpeaks) > 0:
                self.all_peaks_artist.set_data(self.allpeaks[:,0], peak_power)
            else:
                self.all_peaks_artist.set_data([], [])
        if self.power_artist is None:
            self.power_artist, = self.axp.plot(self.freqs, self.power, color='tab:blue', zorder=2)
        else:
            self.power_artist.set_data(self.freqs, self.power)

        # power spectrum of envelope:
        self.envfresolution=1.0
        nfft = int(np.round(2**(np.floor(np.log(self.rate/self.envfresolution) / np.log(2.0)) + 1.0)))
        if nfft < 16:
            self.envfresolution *= 0.5
            nfft = 16
        t00 = t0
        t11 = t1
        w = t11-t00
        minw = int(nfft*(cfg.value('minPSDAverages')+1)//2)
        if t11-t00 < minw:
            w = minw
            t11 = t00 + w
        if t11 >= len(self.envelope):
            t11 = len(self.envelope)
            t00 = t11 - w
        if t00 < 0:
            t00 = 0
            t11 = w
        self.envpower, self.envfreqs = ml.psd(self.envelope[t00:t11], NFFT=nfft, noverlap=nfft//2, Fs=self.rate, detrend=ml.detrend_mean)
        self.envpeaks, _, _ = psd_peaks(self.envfreqs, self.envpower, cfg)
        self.axpe.set_xlim(0.0, 100.0)
        self.axpe.set_xlabel('Frequency [Hz]')
        if self.envpower_label is None:
            self.envpower_label = self.axpe.set_ylabel('Envelope power')
        if self.decibel:
            env_peak_power = self.envpeaks[:,1]
            self.envpower = 10.0*np.log10(self.envpower)
            pmin = np.min(self.envpower[self.envfreqs<100.0])
            pmin = np.floor(pmin/10.0)*10.0
            pmax = np.max(self.envpower[self.envfreqs<100.0])
            pmax = np.ceil(pmax/10.0)*10.0
            doty = pmax-5.0
            self.envpower_label.set_text('Envelope power [dB]')
            self.axpe.set_ylim(pmin, pmax)
        else:
            env_peak_power = 10.0**(0.1*self.envpeaks[:,1])
            pmax = np.max(self.envpower[self.envfreqs<100.0])
            doty = pmax
            pmax *= 1.1
            self.envpower_label.set_text('Envelope power')
            self.axpe.set_ylim(0.0, pmax)
        if self.envpower_peaks_artist is None:
            if len(self.allpeaks) > 0:
                self.envpower_peaks_artist, = self.axpe.plot(self.envpeaks[:,0], env_peak_power,
                                                       'o', color='tab:blue')
            else:
                self.envpower_peaks_artist, = self.axpe.plot([], [], 'o', color='tab:blue')
        else:
            if len(self.allpeaks) > 0:
                self.envpower_peaks_artist.set_data(self.envpeaks[:,0], env_peak_power)
            else:
                self.envpower_peaks_artist.set_data([], [])
        if self.envpower_artist is None:
            self.envpower_artist, = self.axpe.plot(self.envfreqs, self.envpower, color='tab:red', zorder=2)
        else:
            self.envpower_artist.set_data(self.envfreqs, self.envpower)
        self.envcutoff_artist.set_text('cutoff={:.0f} Hz'.format(self.envcutofffreq))
        
        if draw:
            self.fig.canvas.draw()
                 
    def keypress(self, event):
        # print 'pressed', event.key
        if event.key in '+=X':
            if self.twindow*self.rate > 20:
                self.twindow *= 0.5
                self.update_plots()
        elif event.key in '-x':
            if self.twindow < len(self.data)/self.rate:
                self.twindow *= 2.0
                self.update_plots()
        elif event.key == 'pagedown':
            if self.toffset + 0.5*self.twindow < len(self.data)/self.rate:
                self.toffset += 0.5*self.twindow
                self.update_plots()
        elif event.key == 'pageup':
            if self.toffset > 0:
                self.toffset -= 0.5*self.twindow
                if self.toffset < 0.0:
                    self.toffset = 0.0
                self.update_plots()
        elif event.key == 'ctrl+pagedown':
            if self.toffset + 5.0*self.twindow < len(self.data)/self.rate:
                self.toffset += 5.0*self.twindow
                self.update_plots()
        elif event.key == 'ctrl+pageup':
            if self.toffset > 0:
                self.toffset -= 5.0*self.twindow
                if self.toffset < 0.0:
                    self.toffset = 0.0
                self.update_plots()
        elif event.key == 'down':
            if self.toffset + self.twindow < len(self.data)/self.rate:
                self.toffset += 0.05*self.twindow
                self.update_plots()
        elif event.key == 'up':
            if self.toffset > 0.0:
                self.toffset -= 0.05*self.twindow
                if self.toffset < 0.0:
                    self.toffset = 0.0
                self.update_plots()
        elif event.key == 'home':
            if self.toffset > 0.0:
                self.toffset = 0.0
                self.update_plots()
        elif event.key == 'end':
            toffs = np.floor(len(self.data)/self.rate / self.twindow) * self.twindow
            if self.toffset < toffs:
                self.toffset = toffs
                self.update_plots()
        elif event.key == 'y':
            h = self.ymax - self.ymin
            c = 0.5*(self.ymax + self.ymin)
            self.ymin = c-h
            self.ymax = c+h
            self.axt.set_ylim(self.ymin, self.ymax)
            self.fig.canvas.draw()
        elif event.key == 'Y':
            h = 0.25*(self.ymax - self.ymin)
            c = 0.5*(self.ymax + self.ymin)
            self.ymin = c-h
            self.ymax = c+h
            self.axt.set_ylim(self.ymin, self.ymax)
            self.fig.canvas.draw()
        elif event.key == 'v':
            t0 = int(np.round(self.toffset*self.rate))
            t1 = int(np.round((self.toffset+self.twindow)*self.rate))
            min = np.min(self.data[t0:t1])
            max = np.max(self.data[t0:t1])
            h = 0.5*(max - min)
            c = 0.5*(max + min)
            self.ymin = c-h
            self.ymax = c+h
            self.axt.set_ylim(self.ymin, self.ymax)
            self.fig.canvas.draw()
        elif event.key == 'V':
            self.ymin = -1.0
            self.ymax = +1.0
            self.axt.set_ylim(self.ymin, self.ymax)
            self.fig.canvas.draw()
        elif event.key == 'left':
            if self.fmin > 0.0:
                fwidth = self.fmax-self.fmin
                self.fmin -= 0.5*fwidth
                self.fmax -= 0.5*fwidth
                if self.fmin < 0.0:
                    self.fmin = 0.0
                    self.fmax = fwidth
                self.axs.set_ylim(self.fmin, self.fmax)
                self.axp.set_xlim(self.fmin, self.fmax)
                self.fig.canvas.draw()
        elif event.key == 'right':
            if self.fmax < 0.5*self.rate:
                fwidth = self.fmax-self.fmin
                self.fmin += 0.5*fwidth
                self.fmax += 0.5*fwidth
                self.axs.set_ylim(self.fmin, self.fmax)
                self.axp.set_xlim(self.fmin, self.fmax)
                self.fig.canvas.draw()
        elif event.key == 'ctrl+left':
            if self.fmin > 0.0:
                fwidth = self.fmax-self.fmin
                self.fmin = 0.0
                self.fmax = fwidth
                self.axs.set_ylim(self.fmin, self.fmax)
                self.axp.set_xlim(self.fmin, self.fmax)
                self.fig.canvas.draw()
        elif event.key == 'ctrl+right':
            if self.fmax < 0.5*self.rate:
                fwidth = self.fmax-self.fmin
                fm = 0.5*self.rate
                self.fmax = np.ceil(fm/fwidth)*fwidth
                self.fmin = self.fmax - fwidth
                if self.fmin < 0.0:
                    self.fmin = 0.0
                    self.fmax = fwidth
                self.axs.set_ylim(self.fmin, self.fmax)
                self.axp.set_xlim(self.fmin, self.fmax)
                self.fig.canvas.draw()
        elif event.key in 'e':
            self.show_envelope = not self.show_envelope
            self.envelope_artist.set_visible(self.show_envelope)
            self.fig.canvas.draw()
        elif event.key in 'C':
            self.envcutofffreq *= 1.5
            self.envelope = envelope(self.data, self.rate, self.envcutofffreq)
            self.update_plots()
        elif event.key in 'c':
            self.envcutofffreq /= 1.5
            self.envelope = envelope(self.data, self.rate, self.envcutofffreq)
            self.update_plots()
        elif event.key in 'T':
            self.envthreshfac *= 1.5
        elif event.key in 't':
            self.envthreshfac /= 1.5
        elif event.key in 'f':
            if self.fmax < 0.5*self.rate or self.fmin > 0.0:
                fwidth = self.fmax-self.fmin
                if self.fmax < 0.5*self.rate:
                    self.fmax = self.fmin + 2.0*fwidth
                elif self.fmin > 0.0:
                    self.fmin = self.fmax - 2.0*fwidth
                    if self.fmin < 0.0:
                        self.fmin = 0.0
                        self.fmax = 2.0*fwidth
                self.axs.set_ylim(self.fmin, self.fmax)
                self.axp.set_xlim(self.fmin, self.fmax)
                self.fig.canvas.draw()
        elif event.key in 'F':
            if self.fmax - self.fmin > 1.0:
                fwidth = self.fmax-self.fmin
                self.fmax = self.fmin + 0.5*fwidth
                self.axs.set_ylim(self.fmin, self.fmax)
                self.axp.set_xlim(self.fmin, self.fmax)
                self.fig.canvas.draw()
        elif event.key in 'r':
            if self.fresolution < 10000.0:
                self.fresolution *= 2.0
                self.update_plots()
        elif event.key in 'R':
            if 1.0/self.fresolution < self.time[-1]:
                self.fresolution *= 0.5
                self.update_plots()
        elif event.key in 'd':
            self.decibel = not self.decibel
            self.update_plots()
        elif event.key in 'm':
            if cfg.value('mainsFreq') == 0.0:
                cfg.set('mainsFreq', self.mains_freq)
            else:
                cfg.set('mainsFreq', 0.0)
            self.update_plots()
        elif event.key == 'escape':
            self.remove_peak_annotation()
            self.remove_env_peak_annotation()
            self.fig.canvas.draw()
        elif event.key in 'h':
            self.help = not self.help
            for ht in self.helptext:
                ht.set_visible(self.help)
            self.fig.canvas.draw()
        elif event.key in 'w':
            self.plot_waveform()
        elif event.key in 'W':
            self.plot_powerspec()
        elif event.key in 's':
            self.save_segment()
        elif event.key in 'S':
            self.save_powerspec()
        elif event.key in 'E':
            self.save_envelope()
            self.save_envelope_powerspec()
        elif event.key in 'p':
            self.play_segment()
        elif event.key in 'P':
            self.play_all()

    def buttonpress(self, event):
        # print 'mouse pressed', event.button, event.key, event.step
        if event.inaxes == self.axp:
            if event.key == 'shift' or event.key == 'control':
                # show next or previous harmonic:
                if event.key == 'shift':
                    if event.button == 1:
                        ftarget = event.xdata/2.0
                    elif event.button == 3:
                        ftarget = event.xdata*2.0
                else:
                    if event.button == 1:
                        ftarget = event.xdata/1.5
                    elif event.button == 3:
                        ftarget = event.xdata*1.5
                foffs = event.xdata - self.fmin
                fwidth = self.fmax - self.fmin
                self.fmin = ftarget - foffs
                self.fmax = self.fmin + fwidth
                self.axs.set_ylim(self.fmin, self.fmax)
                self.axp.set_xlim(self.fmin, self.fmax)
                self.fig.canvas.draw()
            else:
                # put label on peak
                self.remove_peak_annotation()
                # find closest peak:
                fwidth = self.fmax - self.fmin
                peakdist = np.abs(self.allpeaks[:,0]-event.xdata)
                inx = np.argmin(peakdist)
                if peakdist[inx] < 0.005*fwidth:
                    peak = self.allpeaks[inx,:]
                    self.annotate_peak(peak)
                    self.fig.canvas.draw()
                else:
                    self.fig.canvas.draw()
        if event.inaxes == self.axpe:
            # put label on peak
            self.remove_env_peak_annotation()
            # find closest peak:
            fwidth = 100.0
            peakdist = np.abs(self.envpeaks[:,0]-event.xdata)
            inx = np.argmin(peakdist)
            if peakdist[inx] < 0.05*fwidth:
                peak = self.envpeaks[inx,:]
                self.annotate_env_peak(peak)
                self.fig.canvas.draw()
            else:
                self.fig.canvas.draw()

    def onpick(self, event):
        print('pick')

    def analyse_envelopepeaks(self, tmin, tmax):
        t0 = int(tmin*self.rate)
        t1 = int(tmax*self.rate)
        threshold = self.envthreshfac*np.diff(np.quantile(self.envelope[t0:t1], [0.05, 0.95]))[0]
        peaks, _ = detect_peaks_fixed(self.envelope[t0:t1], threshold)
        npeaks = len(peaks)
        rate = 0.0
        interval = 0.0
        if npeaks > 1:
            peaktimes = self.time[t0:t1][peaks]
            if self.envpeaks_artist is None:
                self.envpeaks_artist, = self.axt.plot(peaktimes, self.envelope[t0+peaks], 'o', color='tab:orange')
            else:
                self.envpeaks_artist.set_data(peaktimes, self.envelope[t0+peaks])
            rate = (npeaks-1.0)/(peaktimes[-1]-peaktimes[0])
            interval = 1.0/rate
        return npeaks, interval, rate

    def analyse_trace(self, tmin, tmax):
        t0 = int(tmin*self.rate)
        t1 = int(tmax*self.rate)
        npeaks, pinterval, prate = self.analyse_envelopepeaks(tmin, tmax)
        peak_freq = 0.001*self.allpeaks[np.argmax(self.allpeaks[:,1]),0]
        peak_width = 0.001*self.allpeaks[np.argmax(self.allpeaks[:,1]),3]
        print('\t'.join([ '{:10s}'.format(x) for x in [ "# width [s]", "trace mean", "trace std", "peak freq [kHz]", "peak width [kHz]", "env mean", "env std", "env peaks", "env T [s]", "env rate [Hz]" ] ]))
        print('\t'.join('{:10.4f}'.format(x) for x in [tmax-tmin, np.mean(self.data[t0:t1]), np.std(self.data[t0:t1]), peak_freq, peak_width, np.mean(self.envelope[t0:t1]), np.std(self.envelope[t0:t1]), npeaks, pinterval, prate]))
        if self.analysis_file is None:
            name = os.path.splitext(self.filename)[0]
            if self.channel > 0:
                datafile = '{name}-{channel:d}-data.csv'.format(
                    name=name, channel=self.channel)
            else:
                datafile = '{name}-data.csv'.format(name=name)
            self.analysis_file = open(os.path.join(self.filepath, datafile), 'w')
            self.analysis_file.write(';'.join([ '{:s}'.format(x) for x in [ "width [s]", "trace mean", "trace std", "peak freq [kHz]", "peak width [kHz]", "env mean", "env std", "env peaks", "env T [s]", "env rate [Hz]" ] ]) + '\n')
            print('saved selected data to: %s' % datafile)
        self.analysis_file.write(';'.join('{:.4f}'.format(x) for x in [ tmax-tmin, np.mean(self.data[t0:t1]), np.std(self.data[t0:t1]), peak_freq, peak_width, np.mean(self.envelope[t0:t1]), np.std(self.envelope[t0:t1]), npeaks, pinterval, prate ]) + '\n')
        self.analysis_file.flush()
            

    def plot_waveform(self):
        fig, ax = plt.subplots()
        name = os.path.splitext(self.filename)[0]
        if self.channel > 0:
            ax.set_title('{filename} channel={channel:d}'.format(
                filename=self.filename, channel=self.channel))
            figfile = '{name}-{channel:d}-{time:.4g}s-waveform.png'.format(
                name=name, channel=self.channel, time=self.toffset)
        else:
            ax.set_title(self.filename)
            figfile = '{name}-{time:.4g}s-waveform.png'.format(
                name=name, time=self.toffset)
        t0 = int(np.round(self.toffset*self.rate))
        t1 = int(np.round((self.toffset+self.twindow)*self.rate))
        if self.twindow < 1.0:
            ax.set_xlabel('Time [ms]')
            ax.set_xlim(1000.0*self.toffset,
                         1000.0*(self.toffset+self.twindow))
            ax.plot(1000.0*self.time[t0:t1], self.data[t0:t1], 'color=tab:blue')
            if self.show_envelope:
                ax.plot(1000.0*self.time[t0:t1], self.envelope[t0:t1], color='tab:red')
        else:
            ax.set_xlabel('Time [s]')
            ax.set_xlim(self.toffset, self.toffset+self.twindow)
            ax.plot(self.time[t0:t1], self.data[t0:t1], color='tab:blue')
            if self.show_envelope:
                ax.plot(self.time[t0:t1], self.envelope[t0:t1], color='tab:red')
        ax.set_ylabel('Amplitude [{:s}]'.format(self.unit))
        fig.tight_layout()
        # on linux the following is not what you want! You want to save this into the current working directory!!!
        fig.savefig(os.path.join(self.filepath, figfile))
        fig.clear()
        plt.close(fig)
        print('saved waveform figure to: %s' % figfile)

    def plot_powerspec(self):
        fig, ax = plt.subplots()
        name = os.path.splitext(self.filename)[0]
        if self.channel > 0:
            ax.set_title('{filename} channel={channel:d}'.format(
                filename=self.filename, channel=self.channel))
            figfile = '{name}-{channel:d}-{time:.4g}s-powerspec.png'.format(
                name=name, channel=self.channel, time=self.toffset)
        else:
            ax.set_title(self.filename)
            figfile = '{name}-{time:.4g}s-powerspec.png'.format(
                name=name, time=self.toffset)
        ax.set_xlabel('Frequency [Hz]')
        ax.set_xlim(self.fmin, self.fmax)
        if self.decibel:
            ax.set_ylabel('Signal power [dB]')
        else:
            ax.set_ylabel('Signal power')
        ax.plot(self.freqs, self.power, color='tab:blue')
        fig.tight_layout()
        # on linux the following is not what you want! You want to save this into the current working directory!!!
        fig.savefig(os.path.join(self.filepath, figfile))
        fig.clear()
        plt.close(fig)
        print('saved power spectrum figure to: %s' % figfile)

        
    def save_segment(self):
        t0s = int(np.round(self.toffset))
        t1s = int(np.round(self.toffset + self.twindow))
        t0 = int(np.round(self.toffset * self.rate))
        t1 = int(np.round((self.toffset + self.twindow) * self.rate))
        savedata = 1.0 * self.data[t0:t1]
        filename = self.filename.split('.')[0]
        if self.channel > 0:
            segmentfilename = '{name}-{channel:d}-{time0:.4g}s-{time1:.4g}s.wav'.format(
                name=filename, time0=t0s, time1=t1s)
        else:
            segmentfilename = '{name}-{time0:.4g}s-{time1:.4g}s.wav'.format(
                name=filename, time0=t0s, time1=t1s)
        write_audio(segmentfilename, savedata, self.rate)
        print('saved segment to: ' , segmentfilename)
    

    def save_powerspec(self):
        t0s = int(np.round(self.toffset))
        t1s = int(np.round(self.toffset + self.twindow))
        t0 = int(np.round(self.toffset * self.rate))
        t1 = int(np.round((self.toffset + self.twindow) * self.rate))
        filename = self.filename.split('.')[0]
        if self.channel > 0:
            filename = '{name}-{channel:d}-{time0:.4g}s-{time1:.4g}s-powerspec.csv'.format(
                name=filename, time0=t0s, time1=t1s)
        else:
            filename = '{name}-{time0:.4g}s-{time1:.4g}s-powerspec.csv'.format(
                name=filename, time0=t0s, time1=t1s)
        punit = 'x^2/Hz'
        if self.decibel:
            punit = 'dB'
        with open(filename, 'w') as df:
            df.write('# {:<7s}\t{:s}\n'.format('freq', 'power'))
            df.write('# {:<7s}\t{:s}\n'.format('Hz', punit))
            for f, p in zip(self.freqs, self.power):
                df.write('{:9.2f}\t{:g}\n'.format(f, p))
        print('saved power spectrum data to: %s' % filename)

        
    def save_envelope(self):
        t0s = int(np.round(self.toffset))
        t1s = int(np.round(self.toffset + self.twindow))
        t0 = int(np.round(self.toffset * self.rate))
        t1 = int(np.round((self.toffset + self.twindow) * self.rate))
        savedata = 1.0 * self.envelope[t0:t1]
        filename = self.filename.split('.')[0]
        if self.channel > 0:
            segmentfilename = '{name}-{channel:d}-{time0:.4g}s-{time1:.4g}s-envelope.wav'.format(
                name=filename, time0=t0s, time1=t1s)
        else:
            segmentfilename = '{name}-{time0:.4g}s-{time1:.4g}s-envelope.wav'.format(
                name=filename, time0=t0s, time1=t1s)
        write_audio(segmentfilename, savedata, self.rate)
        print('saved envelope to: ' , segmentfilename)

    def save_envelope_powerspec(self):
        t0s = int(np.round(self.toffset))
        t1s = int(np.round(self.toffset + self.twindow))
        t0 = int(np.round(self.toffset * self.rate))
        t1 = int(np.round((self.toffset + self.twindow) * self.rate))
        filename = self.filename.split('.')[0]
        if self.channel > 0:
            filename = '{name}-{channel:d}-{time0:.4g}s-{time1:.4g}s-envelope-powerspec.csv'.format(
                name=filename, time0=t0s, time1=t1s)
        else:
            filename = '{name}-{time0:.4g}s-{time1:.4g}s-envelope-powerspec.csv'.format(
                name=filename, time0=t0s, time1=t1s)
        punit = 'x^2/Hz'
        if self.decibel:
            punit = 'dB'
        with open(filename, 'w') as df:
            df.write('# {:<7s}\t{:s}\n'.format('freq', 'power'))
            df.write('# {:<7s}\t{:s}\n'.format('Hz', punit))
            for f, p in zip(self.envfreqs, self.envpower):
                df.write('{:9.2f}\t{:g}\n'.format(f, p))
        print('saved power spectrum of envelope to: %s' % filename)
        

    def play_segment(self):
        t0 = int(np.round(self.toffset*self.rate))
        t1 = int(np.round((self.toffset+self.twindow)*self.rate))
        playdata = 1.0*self.data[t0:t1]
        fade(playdata, self.rate, 0.1)
        self.audio.play(playdata, self.rate, blocking=False)
        
    def play_all(self):
        self.audio.play(self.data, self.rate, blocking=False)
                    

def main(cargs):
    # config file name:
    cfgfile = __package__ + '.cfg'

    # command line arguments:
    parser = argparse.ArgumentParser(description='Display waveform, spectrogram, power spectrum, envelope, and envelope spectrum of time series data.', epilog=f'version {__version__} by Jan Benda (2015-{__year__})')
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument('-v', action='count', dest='verbose',
                        help='print debug information')
    parser.add_argument('-c', '--save-config', nargs='?', default='', const=cfgfile, type=str, metavar='cfgfile',
                        help='save configuration to file cfgfile (defaults to {0})'.format(cfgfile))
    parser.add_argument('-f', dest='high_pass', type=float, metavar='FREQ', default=None,
                        help='cutoff frequency of highpass filter in Hz')
    parser.add_argument('-l', dest='low_pass', type=float, metavar='FREQ', default=None,
                        help='cutoff frequency of lowpass filter in Hz')
    parser.add_argument('file', nargs='?', default='', type=str, help='name of the file with the time series data')
    parser.add_argument('channel', nargs='?', default=0, type=int, help='channel to be displayed')
    args = parser.parse_args(cargs)

    # set configuration from command line:
    if args.verbose != None:
        cfg.set('verboseLevel', args.verbose)

    # load configuration file:
    cfg.load_files(cfgfile, args.file, 3, cfg.value('verboseLevel'))
    
    # save configuration:
    if args.save_config:
        cfg.load_files(cfgfile, args.file, 3, cfg.value('verboseLevel'))
        ext = os.path.splitext(cfgfile)[1]
        if ext != os.extsep + 'cfg':
            print('configuration file name must have .cfg as extension!')
        else:
            print('write configuration to %s ...' % cfgfile)
            cfg.dump(cfgfile)
        exit()

    # load data:
    filepath = args.file
    channel = args.channel
    filename = os.path.basename(filepath)
    ext = os.path.splitext(filename)[1]
    data, rate = load_audio(filepath)
    data = data[:,channel]
    if not args.high_pass is None:
        if not args.low_pass is None:
            data = bandpass_filter(data, rate, args.high_pass, args.low_pass)
        else:
            data = highpass_filter(data, rate, args.high_pass)
    unit = 'a.u.'
    
    # plot:
    sp = SignalPlot(data, rate, unit, filename, channel, os.path.dirname(filepath))


def run():
    main(sys.argv[1:])

    
if __name__ == '__main__':
    run()



import numpy as np
try:
    from PyQt5.QtCore import Signal
except ImportError:
    from PyQt5.QtCore import pyqtSignal as Signal
from PyQt5.QtCore import Qt, QRectF
from PyQt5.QtGui import QTransform
import pyqtgraph as pg


class SelectViewBox(pg.ViewBox):


    sigSelectedRegion = Signal(object, object, object)


    def __init__(self, channel, *args, **kwargs):
        pg.ViewBox.__init__(self, *args, **kwargs)
        self.setMouseMode(pg.ViewBox.RectMode)
        self.rbScaleBox.setPen(pg.mkPen((200, 200, 200), width=1))
        self.rbScaleBox.setBrush(pg.mkBrush(200, 200, 200, 100))
        self.channel = channel


    def keyPressEvent(self, ev):
        ev.ignore()
        
        
    def mouseDragEvent(self, ev, axis=None):
        ## if axis is specified, event will only affect that axis.
        ev.accept()  ## we accept all buttons

        pos = ev.pos()
        lastPos = ev.lastPos()
        dif = pos - lastPos
        dif = dif * -1

        ## Ignore axes if mouse is disabled
        mouseEnabled = np.array(self.state['mouseEnabled'], dtype=np.float64)
        mask = mouseEnabled.copy()
        if axis is not None:
            mask[1-axis] = 0.0

        ## Scale or translate based on mouse button
        if ev.button() in [Qt.MouseButton.LeftButton, Qt.MouseButton.MiddleButton]:
            if self.state['mouseMode'] == pg.ViewBox.RectMode and axis is None:
                if ev.isFinish():
                    # This is the final move in the drag; change the view scale now
                    rect = QRectF(pg.Point(ev.buttonDownPos(ev.button())), pg.Point(pos))
                    rect = self.childGroup.mapRectFromParent(rect) # in data coordinates
                    self.sigSelectedRegion.emit(self.channel, self, rect)
                else:
                    ## update shape of scale box
                    self.updateScaleBox(ev.buttonDownPos(), ev.pos())
            else:
                tr = self.childGroup.transform()
                tr = pg.functions.invertQTransform(tr)
                tr = tr.map(dif*mask) - tr.map(pg.Point(0,0))

                x = tr.x() if mask[0] == 1 else None
                y = tr.y() if mask[1] == 1 else None

                self._resetTarget()
                if x is not None or y is not None:
                    self.translateBy(x=x, y=y)
                self.sigRangeChangedManually.emit(self.state['mouseEnabled'])
                if ev.isFinish():
                    self.add_region(self.viewRect())
        elif ev.button() & Qt.MouseButton.RightButton:
            #print "vb.rightDrag"
            if self.state['aspectLocked'] is not False:
                mask[0] = 0

            dif = ev.screenPos() - ev.lastScreenPos()
            dif = np.array([dif.x(), dif.y()])
            dif[0] *= -1
            s = ((mask * 0.02) + 1) ** dif

            tr = self.childGroup.transform()
            tr = pg.functions.invertQTransform(tr)

            x = s[0] if mouseEnabled[0] == 1 else None
            y = s[1] if mouseEnabled[1] == 1 else None

            center = pg.Point(tr.map(ev.buttonDownPos(Qt.MouseButton.RightButton)))
            self._resetTarget()
            self.scaleBy(x=x, y=y, center=center)
            self.sigRangeChangedManually.emit(self.state['mouseEnabled'])
            if ev.isFinish():
                self.add_region(self.viewRect())


    def updateScaleBox(self, p1, p2):
        r = QRectF(p1, p2)
        r = self.childGroup.mapRectFromParent(r)
        self.rbScaleBox.setPos(r.topLeft())
        tr = QTransform.fromScale(r.width(), r.height())
        self.rbScaleBox.setTransform(tr)
        self.rbScaleBox.show()

        
    def hide_region(self):
        self.rbScaleBox.hide()

        
    def add_region(self, rect):
        self.axHistoryPointer += 1
        self.axHistory = self.axHistory[:self.axHistoryPointer] + [rect]

        
    def zoom_region(self, rect):
        self.hide_region()
        self.showAxRect(rect)
        self.add_region(rect)


    def zoom_back(self):
        self.scaleHistory(-1)


    def zoom_forward(self):
        self.scaleHistory(1)


    def zoom_home(self):
        self.scaleHistory(-len(self.axHistory))


    def init_zoom_history(self):
        self.add_region(self.viewRect())



from math import floor, fabs
import numpy as np
from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtWidgets import QGraphicsSimpleTextItem, QApplication
import pyqtgraph as pg
from .traceitem import down_sample_peak


def secs_to_str(time):
    hours = int(time//3600)
    time -= 3600*hours
    mins = int(time//60)
    time -= 60*mins
    secs = int(floor(time))
    time -= secs
    if hours > 0:
        return f'{hours}:{mins:02d}:{secs:02d}'
    elif mins > 0:
        return f'{mins:02d}:{secs:02d}'
    elif secs > 0:
        return f'{secs}.{1000*time:03.0f}s'
    elif time >= 0.01:
        return f'{1000*time:03.0f}ms'
    elif time >= 0.001:
        return f'{1000*time:.2f}ms'
    else:
        return f'{1e6*time:.0f}\u00b5s'


def secs_format(time):
    if time >= 3600.0:
        return 'h:mm:ss'
    elif time >= 60.0:
        return 'mm:ss'
    elif time > 1.0:
        return 's.ms'
    else:
        return 'ms'

    
class FullTracePlot(pg.GraphicsLayoutWidget):

    
    def __init__(self, data, rate, axtraces, *args, **kwargs):
        pg.GraphicsLayoutWidget.__init__(self, *args, **kwargs)

        self.data = data
        self.rate = rate
        self.tmax = len(self.data)/self.rate
        self.axtraces = axtraces
        self.no_signal = False

        self.setBackground(None)
        self.ci.layout.setContentsMargins(0, 0, 0, 0)
        self.ci.layout.setVerticalSpacing(0)

        # for each channel prepare a plot panel:
        xwidth = self.fontMetrics().averageCharWidth()
        self.axs = []
        self.lines = []
        self.regions = []
        self.labels = []
        for c in range(self.data.channels):
            # setup plot panel:
            axt = pg.PlotItem()
            axt.showAxes(True, False)
            axt.getAxis('left').setWidth(8*xwidth)
            axt.getViewBox().setBackgroundColor(None)
            axt.getViewBox().setDefaultPadding(padding=0.0)
            axt.hideButtons()
            axt.setMenuEnabled(False)
            axt.setMouseEnabled(False, False)
            axt.enableAutoRange(False, False)
            axt.setLimits(xMin=0, xMax=self.tmax,
                          minXRange=self.tmax, maxXRange=self.tmax)
            axt.setXRange(0, self.tmax)

            # add region marker:
            region = pg.LinearRegionItem(pen=dict(color='#110353', width=2),
                                         brush=(34, 6, 167, 127),
                                         hoverPen=dict(color='#aa77ff', width=2),
                                         hoverBrush=(34, 6, 167, 255),
                                         movable=True,
                                         swapMode='block')
            region.setZValue(50)
            region.setBounds((0, self.tmax))
            region.setRegion((self.axtraces[c].viewRange()[0]))
            region.sigRegionChanged.connect(self.update_time_range)
            self.axtraces[c].sigXRangeChanged.connect(self.update_region)
            axt.addItem(region)
            self.regions.append(region)

            # add time label:
            label = QGraphicsSimpleTextItem(axt.getAxis('left'))
            label.setToolTip(f'Total duration in {secs_format(self.tmax)}')
            label.setText(secs_to_str(self.tmax))
            label.setPos(int(xwidth), xwidth/2)
            self.labels.append(label)

            # init data:
            max_pixel = QApplication.desktop().screenGeometry().width()
            self.step = max(1, len(self.data)//max_pixel)
            self.index = 0
            self.nblock = int(20.0*self.rate//self.step)*self.step
            self.times = np.arange(0, len(self.data), self.step/2)/self.rate
            self.datas = np.zeros((len(self.times), self.data.channels))
            
            # add data:
            line = pg.PlotDataItem(antialias=True,
                                   pen=dict(color='#2206a7', width=1.1),
                                   skipFiniteCheck=True, autDownsample=False)
            line.setZValue(10)
            axt.addItem(line)
            self.lines.append(line)

            # add zero line:
            zero_line = axt.addLine(y=0, movable=False, pen=dict(color='grey', width=1))
            zero_line.setZValue(20)
            
            self.addItem(axt, row=c, col=0)
            self.axs.append(axt)
            
        QTimer.singleShot(10, self.load_data)


    def load_data(self):
        i = 2*self.index//self.step
        n = min(self.nblock, len(self.data) - self.index)
        buffer = np.zeros((n, self.data.channels))
        self.data.load_buffer(self.index, n, buffer)
        for c in range(self.data.channels):
            data = down_sample_peak(buffer[:,c], self.step)
            self.datas[i:i+len(data), c] = data
            self.lines[c].setData(self.times, self.datas[:,c])
        self.index += n
        if self.index < len(self.data):
            QTimer.singleShot(10, self.load_data)
        else:
            for c in range(self.data.channels):
                ymin = np.min(self.datas[:,c])
                ymax = np.max(self.datas[:,c])
                y = max(fabs(ymin), fabs(ymax))
                self.axs[c].setYRange(-y, y)
                self.axs[c].setLimits(yMin=-y, yMax=y,
                                      minYRange=2*y, maxYRange=2*y)
            self.times = None
            self.datas = None

        
    def update_layout(self, channels, data_height):
        first = True
        for c in range(self.data.channels):
            self.axs[c].setVisible(c in channels)
            if c in channels:
                self.ci.layout.setRowFixedHeight(c, data_height)
                self.labels[c].setVisible(first)
                first = False
            else:
                self.ci.layout.setRowFixedHeight(c, 0)
                self.labels[c].setVisible(False)
        self.setFixedHeight(len(channels)*data_height)


    def update_time_range(self, region):
        if self.no_signal:
            return
        self.no_signal = True
        xmin, xmax = region.getRegion()
        for ax, reg in zip(self.axtraces, self.regions):
            if reg is region:
                ax.setXRange(xmin, xmax)
                break
        self.no_signal = False


    def update_region(self, vbox, x_range):
        for ax, region in zip(self.axtraces, self.regions):
            if ax.getViewBox() is vbox:
                region.setRegion(x_range)
                break

        
    def mousePressEvent(self, ev):
        if ev.button() == Qt.MouseButton.LeftButton:
            for ax, region in zip(self.axs, self.regions):
                pos = ax.getViewBox().mapSceneToView(ev.pos())
                [xmin, xmax], [ymin, ymax] = ax.viewRange()
                if xmin <= pos.x() <= xmax and ymin <= pos.y() <= ymax:
                    dx = (xmax - xmin)/self.width()
                    x = pos.x()
                    xmin, xmax = region.getRegion()
                    if x < xmin-2*dx or x > xmax + 2*dx:
                        dx = xmax - xmin
                        xmin = max(0, x - dx/2)
                        xmax = xmin + dx
                        if xmax > self.tmax:
                            xmin = max(0, xmax - dx)
                        region.setRegion((xmin, xmax))
                        ev.accept()
                        return
                    break
        ev.ignore()
        super().mousePressEvent(ev)



from math import fabs, floor, ceil
import numpy as np
import scipy.signal as sig
from PyQt5.QtWidgets import QApplication
import pyqtgraph as pg

has_numba = False
try:
    from numba import jit
    has_numba = True
except ImportError:
    def jit(*args, **kwargs):
        def decorator_jit(func):
            return func
        return decorator_jit


@jit(nopython=True)
def down_sample_peak(data, step):
    n = len(data)//step
    ddata = np.zeros(2*n)
    for k in range(n):
        dd = data[k*step:(k+1)*step]
        ddata[2*k] = np.min(dd)
        ddata[2*k+1] = np.max(dd)
    return ddata

    
class TraceItem(pg.PlotDataItem):
    
    def __init__(self, data, rate, channel, *args, color='#00ee00', **kwargs):
        self.data = data
        self.rate = rate
        self.channel = channel
        self.step = 1
        self.color = color
        self.ymin = self.data.ampl_min
        self.ymax = self.data.ampl_max
        self.highpass_cutoff = None
        self.lowpass_cutoff = None
        self.sos = None
        
        pg.PlotDataItem.__init__(self, *args, connect='all',
                                 antialias=False, skipFiniteCheck=True,
                                 **kwargs)
        self.setPen(dict(color=self.color, width=2))
        self.setSymbolSize(8)
        self.setSymbolBrush(color=self.color)
        self.setSymbolPen(color=self.color)
        self.setSymbol(None)


    def set_color(self, color):
        self.color = color
        self.setPen(dict(color=self.color, width=2))
        self.setSymbolBrush(color=self.color)
        self.setSymbolPen(color=self.color)

        
    def viewRangeChanged(self):
        self.updateTrace()
    

    def updateTrace(self):
        vb = self.getViewBox()
        if not isinstance(vb, pg.ViewBox):
            return

        # time:
        trange = vb.viewRange()[0]
        start = max(0, int(trange[0]*self.rate))
        stop = min(len(self.data), int(trange[1]*self.rate+1))
        sstop = int(trange[1]*self.rate+1)
        max_pixel = QApplication.desktop().screenGeometry().width()
        self.step = max(1, (sstop - start)//max_pixel)
        if self.step > 1:
            step2 = self.step//2
            self.step = step2*2
            start = int(floor(start/self.step)*self.step)
            stop = int(ceil(stop/self.step)*self.step)
            self.setPen(dict(color=self.color, width=1.1))
            n = (stop-start)//self.step
            data = np.zeros(2*n)
            i = 0
            nb = int(60*self.rate//self.step)*self.step
            for dd in self.data.blocks(nb, 0, start, stop):
                db = dd[:, self.channel]
                if self.sos is not None:
                    # TODO: extend data!!!
                    db = sig.sosfiltfilt(self.sos, db)
                if has_numba:
                    dsd = down_sample_peak(db, self.step)
                    data[i:i+len(dsd)] = dsd
                    i += len(dsd)
                #else:
                #    data = np.array([(np.min(self.data[start+k*self.step:start+(k+1)*self.step, self.channel]), np.max(self.data[start+k*self.step:start+(k+1)*self.step, self.channel])) for k in range(n)]).reshape((-1))
            time = np.arange(start, start + len(data)*step2, step2)/self.rate
            #print('T', time[0], time[-1])
            self.setData(time, data) #, connect='pairs')???
        elif self.step > 1:  # TODO: not used
            # subsample:
            self.setData(np.arange(start, stop, self.step)/self.rate,
                         self.data[start:stop:self.step, self.channel])
            self.setPen(dict(color=self.color, width=1.1))
        else:
            # all data:
            self.setData(np.arange(start, stop)/self.rate,
                         self.data[start:stop, self.channel])
            self.setPen(dict(color=self.color, width=2))
            if max_pixel/(stop - start) >= 10:
                self.setSymbol('o')
            else:
                self.setSymbol(None)


    def get_amplitude(self, x, y, x1=None):
        """Get trace amplitude next to cursor position. """
        idx = int(np.round(x*self.rate))
        step = self.step
        if x1 is not None:
            idx1 = int(np.round(x1*self.rate))
            step = max(1, idx1 - idx)
        if step > 1:
            idx = (idx//step)*step
            data_block = self.data[idx:idx + step, self.channel]
            mini = np.argmin(data_block)
            maxi = np.argmax(data_block)
            amin = data_block[mini]
            amax = data_block[maxi]
            if fabs(y - amax) < fabs(y - amin):
                return (idx+maxi)/self.rate, amax
            else:
                return (idx+mini)/self.rate, amin
        else:
            return idx/self.rate, self.data[idx, self.channel]

        
    def zoom_ampl_in(self):
        h = 0.25*(self.ymax - self.ymin)
        c = 0.5*(self.ymax + self.ymin)
        if h > 1/2**16:
            self.ymin = c - h
            self.ymax = c + h

        
    def zoom_ampl_out(self):
        h = self.ymax - self.ymin
        c = 0.5*(self.ymax + self.ymin)
        self.ymin = c - h
        self.ymax = c + h
        if self.ymax > self.data.ampl_max:
            self.ymax = self.data.ampl_max
            self.ymin = 1 - 2*h
        if self.ymin < self.data.ampl_min:
            self.ymin = self.data.ampl_min
        
        
    def auto_ampl(self, toffset, twindow):
        t0 = int(np.round(toffset * self.rate))
        t1 = int(np.round((toffset + twindow) * self.rate))
        ymin = np.min(self.data[t0:t1, self.channel])
        ymax = np.max(self.data[t0:t1, self.channel])
        h = 0.5*(ymax - ymin)
        
        c = 0.5*(ymax + ymin)
        if h < 1/2**16:
            h = 1/2**16
        self.ymin = c - h
        self.ymax = c + h

        
    def reset_ampl(self):
        self.ymin = self.data.ampl_min
        self.ymax = self.data.ampl_max


    def center_ampl(self):
        dy = self.ymax - self.ymin
        self.ymin = -dy/2
        self.ymax = +dy/2


    def set_filter(self, highpass_cutoff, lowpass_cutoff):
        if highpass_cutoff is not None:
            self.highpass_cutoff = highpass_cutoff
        if lowpass_cutoff is not None:
            self.lowpass_cutoff = lowpass_cutoff
        if self.highpass_cutoff < 1e-8 and \
           self.lowpass_cutoff >= self.rate/2-1e-8:
            self.sos = None
        elif self.highpass_cutoff < 1e-8:
            self.sos = sig.butter(2, self.lowpass_cutoff,
                                  'lowpass', fs=self.rate, output='sos')
        elif self.lowpass_cutoff >= self.rate/2-1e-8:
            self.sos = sig.butter(2, self.highpass_cutoff,
                                  'highpass', fs=self.rate, output='sos')
        else:
            self.sos = sig.butter(2, (self.highpass_cutoff, self.lowpass_cutoff),
                                  'bandpass', fs=self.rate, output='sos')
        self.updateTrace()
        
